import torch
import smplx
import numpy as np
import time
import copy

from .vision3d import projectP
from .losses import SegmentTriangleMeshIntersectionLossDualHands, getLossFnc, compute_pearson_depth_loss
from .temporal_losses import getAccelerationStats, getVelocityStats, getJerkStats
from .bspline_kernel import NeuralSplineFilterDecoder
from .smpl_fx.body_models import SMPLFX
from .maths_utils import avgL2distance, euler_to_6d_pytorch3d
from .utils import PrintColors, createChunkedTemporalIndices, getTemporalIndicesBasedKeys, matrix_to_euler_angles, euler_angles_to_matrix, calculateInitialLogits
from .utils_smplx import getContactWeights

def is_shape(x):
    return x == 'shape' or x == 'betas'

class SMPLXWrapper():
    def __init__(self, NUM_POSES, is_smplfx, smplx_kwargs, smplx_params_init, smplx_cp_params_init, chunk_size, optimize_shape, optimize_hands, optimize_smplx_rt, are_control_points, device, GLOBAL_ORIENT_ZERO_X, GLOBAL_ORIENT_FIX_Z, right_hand_constraints, left_hand_constraints, body_zero_angles_idxs, LRs, USE_SIGMOID_ACTIVATION, spacing, apply_constraints_before_decode, OPTIMIZE_FLAME_SHAPE, body_constraints, is_single_smplx_rt, OPTIMIZE_FACES, IS_BODY_POSE_6D, OPTIMIZE_BODY_POSE):
        self.are_control_points = are_control_points
        self.USE_SIGMOID_ACTIVATION = USE_SIGMOID_ACTIVATION
        if USE_SIGMOID_ACTIVATION:
            print(PrintColors.WARNING + "Using Sigmoid Activation for constraints enforcement seems worse than clamping!" + PrintColors.ENDC)

        self.device = device
        if is_smplfx:
            self.model = SMPLFX(**smplx_kwargs).to(device)
        else:
            self.model = smplx.create(**smplx_kwargs).to(device)
        for param in self.model.parameters():
            param.requires_grad = False

        self.LRs = LRs
        self.is_single_smplx_rt = is_single_smplx_rt
        self.smplx_params = {}
        self.OPTIM_J_REGRESSOR = LRs.get('J_regressor', 0) > 0
        J_regressor_copy = self.model.J_regressor.detach().clone().to(device)
        J_regressor_bool = J_regressor_copy > 0
        self.J_regressor_zero_mask = J_regressor_bool.type(torch.float32)
        if 'J_regressor' in smplx_params_init:
            J_regressor_init = smplx_params_init['J_regressor']
        elif self.OPTIM_J_REGRESSOR:
            J_regressor_init = J_regressor_copy
        else:
            J_regressor_init = None
        if J_regressor_init is None:
            self.J_regressor_log = None
        else:
            self.J_regressor_log = torch.log(J_regressor_init.detach().clone().to(device) + 1e-9).clone().requires_grad_(self.OPTIM_J_REGRESSOR)

        self.constraints = {'eyelid_params': torch.tensor([[0.0, 1.0], [0.0, 1.0]], device=device)}
        if right_hand_constraints is not None:
            assert right_hand_constraints.shape == (15*3, 2)
            self.constraints['right_hand_pose'] = right_hand_constraints.to(device)
        if left_hand_constraints is not None:
            assert left_hand_constraints.shape == (15*3, 2)
            self.constraints['left_hand_pose'] = left_hand_constraints.to(device)
        if GLOBAL_ORIENT_ZERO_X or GLOBAL_ORIENT_FIX_Z:
            self.constraints['global_orient'] = torch.tensor([[-np.pi, np.pi], [-np.pi, np.pi], [-np.pi, np.pi]], device=device)
            if GLOBAL_ORIENT_ZERO_X:
                self.constraints['global_orient'][0,:] = 0.0
            if GLOBAL_ORIENT_FIX_Z:
                self.constraints['global_orient'][2,:] = np.pi
        self.body_zero_angles_idxs = body_zero_angles_idxs
        if body_zero_angles_idxs is not None:
            if body_constraints is not None:
                assert body_constraints.shape == (21*3, 2)
                self.constraints['body_pose'] = body_constraints.to(device)
            else:
                self.constraints['body_pose'] = torch.tensor([[-np.pi, np.pi]], device=device).repeat(21*3,1)
            for idx in body_zero_angles_idxs:
                self.constraints['body_pose'][idx*3:idx*3+3,:] = 0.0

        self.constraints_diff = {k: v[:,1] - v[:,0] for k,v in self.constraints.items()}

        params_dim = {'transl': 3, 'global_orient': 3, 'body_pose': 21*3, 'left_hand_pose': 15*3, 'right_hand_pose': 15*3, 'jaw_pose': 3, 'expression': smplx_kwargs.get('num_expression_coeffs', 10), 'eyelid_params': 2, 'betas': smplx_kwargs.get('num_betas', 10)}
        params_optim = {'transl': optimize_smplx_rt, 'global_orient': optimize_smplx_rt, 'body_pose': OPTIMIZE_BODY_POSE, 'left_hand_pose': optimize_hands, 'right_hand_pose': optimize_hands, 'jaw_pose': OPTIMIZE_FACES, 'expression': OPTIMIZE_FACES, 'eyelid_params': OPTIMIZE_FACES, 'betas': optimize_shape}
        self.is_single = lambda x : is_shape(x) or (x == 'transl' and is_single_smplx_rt) or (x == 'global_orient' and is_single_smplx_rt)
        get_poses = lambda x : 1 if self.is_single(x) else NUM_POSES
        
        if is_smplfx:
            params_dim['shape'] = self.model.FLAME.betas.shape[-1]
            params_optim['shape'] = OPTIMIZE_FLAME_SHAPE

        for key in params_dim.keys():
            if key in smplx_params_init:
                if self.is_single(key):
                    self.smplx_params[key] = smplx_params_init[key].to(device).clone().detach().median(dim=0).values[None].requires_grad_(params_optim[key])
                else:
                    self.smplx_params[key] = smplx_params_init[key].to(device).clone().detach().requires_grad_(params_optim[key])
                assert self.smplx_params[key].ndim == 2
            else:
                self.smplx_params[key] = torch.zeros((get_poses(key), params_dim[key]), device=device, requires_grad=params_optim[key])
            print('smplx params', key, self.smplx_params[key].shape, 'optim:', params_optim[key])
            assert self.smplx_params[key].isnan().any() == False, f'SMPLX param {key} has NaNs!'
            assert self.smplx_params[key].isinf().any() == False, f'SMPLX param {key} has Infs!'

        self.IS_BODY_POSE_6D = IS_BODY_POSE_6D
        if IS_BODY_POSE_6D:
            if not are_control_points:
                print(PrintColors.WARNING + "Using 6D representation is optimal for control points! " + PrintColors.ENDC)
            params_dim['body_pose'] = 21*6
            assert body_constraints is None
            self.constraints.pop('body_pose', None)
            self.smplx_params['body_pose'] = euler_to_6d_pytorch3d(self.smplx_params['body_pose'])
            assert self.smplx_params['body_pose'].isnan().any() == False, f'SMPLX param body_pose has NaNs after 6D conversion!'
            assert self.smplx_params['body_pose'].isinf().any() == False, f'SMPLX param body_pose has Infs after 6D conversion!'
        else:
            if are_control_points and OPTIMIZE_BODY_POSE:
                print(PrintColors.WARNING + "For control points it is better to use 6D representation!" + PrintColors.ENDC)

        with torch.no_grad():
            self.smplx_params['global_orient'][:,2] = np.pi

        self.params_optim = params_optim
        self.optimized_params = {}
        if are_control_points:
            for key in self.smplx_params.keys():
                if (not self.is_single(key)) and params_optim[key]:
                    self.smplx_params[key] = self.smplx_params[key].detach().requires_grad_(False)
                    self.optimized_params[key] = NeuralSplineFilterDecoder(NUM_POSES, params_dim[key], chunk_size, spacing=spacing, constraints=self.constraints.get(key, None) if self.USE_SIGMOID_ACTIVATION else None, init_cp_params=smplx_cp_params_init.get(key, None), init_params=self.smplx_params[key], apply_constraints_before_decode=apply_constraints_before_decode).to(device)
                    self.num_chunks = self.optimized_params[key].num_chunks
                    self.knots_per_chunk = self.optimized_params[key].knots_per_chunk
                    self.total_knots = self.optimized_params[key].total_knots
                    # todo: check the initial decoder estimates so they match the init params!

                    if key in self.constraints and (not self.USE_SIGMOID_ACTIVATION):
                        for chunk_idx in range(self.num_chunks):
                            with torch.no_grad():
                                self.optimized_params[key].chunks[chunk_idx].clamp_(min=self.constraints[key][:,0][None,:,None], max=self.constraints[key][:,1][None,:,None])
                                assert not torch.isnan(self.optimized_params[key].chunks[chunk_idx]).any(), f'NaNs in optimized param {key} chunk {chunk_idx} after clamping!'
                                assert not torch.isinf(self.optimized_params[key].chunks[chunk_idx]).any(), f'Infs in optimized param {key} chunk {chunk_idx} after clamping!'

                    for chunk_idx in range(self.num_chunks):
                        with torch.no_grad():
                            test_decoded, _ = self.optimized_params[key](chunk_idx, get_loss=False)
                            assert not torch.isnan(test_decoded).any(), f'NaNs in optimized param {key} chunk {chunk_idx} test_decoded!'
                            assert not torch.isinf(test_decoded).any(), f'Infs in optimized param {key} chunk {chunk_idx} test_decoded!'

                        diff_init = torch.linalg.norm(test_decoded - self.smplx_params[key][chunk_idx*self.knots_per_chunk*spacing:(chunk_idx+1)*self.knots_per_chunk*spacing], dim=-1)
                        print('diff mean', diff_init.mean().item(), 'max', diff_init.max().item(), 'median', diff_init.median().item(), 'for key', key, 'chunk', chunk_idx)
                        # if key == 'body_pose':
                        #     with torch.no_grad():
                        #         smplx_out1 = self.model(**{key: test_decoded})
                        #         smplx_out2 = self.model(**{key: self.smplx_params[key][chunk_idx*self.knots_per_chunk*spacing:(chunk_idx+1)*self.knots_per_chunk*spacing]})
                        #         plotMesh([smplx_out1['vertices'][-14]], self.model.faces, title='check decoded vs init for '+key+' chunk '+str(chunk_idx), show=False, elev=90, azim=-90)
                        #         plotMesh([smplx_out2['vertices'][-14]], self.model.faces, title='check init for '+key+' chunk '+str(chunk_idx), show=True, elev=90, azim=-90)
                        #         pdb.set_trace()

        else:
            self.num_chunks = (NUM_POSES + chunk_size - 1) // chunk_size
            for key in self.smplx_params.keys():
                if (not self.is_single(key)) and params_optim[key]:
                    self.smplx_params[key] = self.smplx_params[key].detach().requires_grad_(False)
                    if key in self.constraints and self.USE_SIGMOID_ACTIVATION:
                        self.smplx_params[key] = calculateInitialLogits(self.smplx_params[key].clone(), self.constraints[key])
                        assert not torch.isnan(self.smplx_params[key]).any()
                        assert not torch.isinf(self.smplx_params[key]).any()
                    self.optimized_params[key] = torch.chunk(self.smplx_params[key].clone(), self.num_chunks, dim=0)
                    self.optimized_params[key] = [p.requires_grad_(True) for p in self.optimized_params[key]]

        for key in self.optimized_params.keys():
            # print('POP parameter', key)
            self.smplx_params.pop(key, None)
        
        for key in self.smplx_params.keys():
            if not self.is_single(key):
                # split into chunks:
                if are_control_points:
                    raise NotImplementedError("Chunking not implemented for control points optimization yet!")
                else:
                    self.smplx_params[key] = torch.chunk(self.smplx_params[key], self.num_chunks, dim=0)

    def constraints_activation(self, x, param_key):
        if (param_key not in self.constraints) or (not self.USE_SIGMOID_ACTIVATION) or (not self.params_optim[param_key]):
            return x
        return torch.sigmoid(x) * self.constraints_diff[param_key] + self.constraints[param_key][:,0]

    def activateJRegressor(self):
        return torch.nn.functional.softmax(self.J_regressor_log * self.J_regressor_zero_mask + (1 - self.J_regressor_zero_mask) * (-1e9), dim=-1)

    def getOptimizableParamsList(self):
        params_list = []
        if self.OPTIM_J_REGRESSOR:
            print(f"Optimizing J_regressor with lr {self.LRs['J_regressor']} device {self.J_regressor_log.device}")
            params_list.append({'params': self.J_regressor_log, 'lr': self.LRs['J_regressor'], 'name': 'J_regressor_log'})
        for key in self.smplx_params.keys():
            if self.params_optim[key]:
                print(f"Optimizing SMPLX param: {key} with lr {self.LRs.get(key, self.LRs['lr'])} device {self.smplx_params[key].device}")
                params_list.append({'params': self.smplx_params[key], 'lr': self.LRs.get(key, self.LRs['lr']), 'name': key})
        for key in self.optimized_params.keys():
            if isinstance(self.optimized_params[key], NeuralSplineFilterDecoder):
                print(f"Optimizing SMPLX control points param: {key} with lr {self.LRs.get(key, self.LRs['lr'])} device {self.optimized_params[key].chunks[0].device}")
                params_list.append({'params': self.optimized_params[key].parameters(), 'lr': self.LRs.get(key, self.LRs['lr']), 'name': key})
            else:
                print(f"Optimizing SMPLX chunked param: {key} with lr {self.LRs.get(key, self.LRs['lr'])} device {self.optimized_params[key][0].device}")
                params_list.append({'params': self.optimized_params[key], 'lr': self.LRs.get(key, self.LRs['lr']), 'name': key})
        return params_list

    def apply_constraints(self, chunk_idx):
        if not self.USE_SIGMOID_ACTIVATION:
            with torch.no_grad():
                for key in self.constraints.keys():
                    if key in self.optimized_params.keys():
                        if self.are_control_points:
                            # chunks are (1, D, C) and constraints are (D), so unsqueeze accordingly to 1 x D x 1
                            self.optimized_params[key].chunks[chunk_idx].clamp_(min=self.constraints[key][:,0][None,:,None], max=self.constraints[key][:,1][None,:,None])
                            if chunk_idx-1 >= 0:
                                self.optimized_params[key].chunks[chunk_idx-1].clamp_(min=self.constraints[key][:,0][None,:,None], max=self.constraints[key][:,1][None,:,None])
                            elif chunk_idx+1 < self.num_chunks:
                                self.optimized_params[key].chunks[chunk_idx+1].clamp_(min=self.constraints[key][:,0][None,:,None], max=self.constraints[key][:,1][None,:,None])
                        else:
                            self.optimized_params[key][chunk_idx].clamp_(min=self.constraints[key][:,0][None], max=self.constraints[key][:,1][None])
                    else:
                        if key in self.smplx_params.keys() and self.params_optim[key]:
                            self.smplx_params[key].clamp_(min=self.constraints[key][:,0][None], max=self.constraints[key][:,1][None])
                if self.IS_BODY_POSE_6D and self.body_zero_angles_idxs is not None:
                    if self.are_control_points:
                        reshaped = self.optimized_params['body_pose'].chunks[chunk_idx].reshape(1, 21, 6, -1)
                        reshaped[:, self.body_zero_angles_idxs] = torch.tensor([1.0, 0.0, 0.0, 0.0, 1.0, 0.0], device=self.device)[None,None,:,None]
                        if chunk_idx-1 >= 0:
                            reshaped1 = self.optimized_params['body_pose'].chunks[chunk_idx-1].reshape(1, 21, 6, -1)
                            reshaped1[:, self.body_zero_angles_idxs] = torch.tensor([1.0, 0.0, 0.0, 0.0, 1.0, 0.0], device=self.device)[None,None,:,None]
                        elif chunk_idx+1 < self.num_chunks:
                            reshaped2 = self.optimized_params['body_pose'].chunks[chunk_idx+1].reshape(1, 21, 6, -1)
                            reshaped2[:, self.body_zero_angles_idxs] = torch.tensor([1.0, 0.0, 0.0, 0.0, 1.0, 0.0], device=self.device)[None,None,:,None]
                    else:
                        reshaped = self.optimized_params['body_pose'][chunk_idx].reshape(-1, 21, 6)
                        reshaped[:, self.body_zero_angles_idxs] = torch.tensor([1.0, 0.0, 0.0, 0.0, 1.0, 0.0], device=self.device)[None,None,:]

    def getParams(self):
        smplx_params = self.getSMPLXParams(chunk_idx=-1, get_loss=False)[0]
        for key in smplx_params.keys():
            smplx_params[key] = smplx_params[key].detach().clone()

        if self.are_control_points:
            optimized_params = {key: copy.deepcopy(self.optimized_params[key].state_dict()) for key in self.optimized_params.keys()}
            return smplx_params, optimized_params
        else:
            return smplx_params, {}

    def getSMPLXParams(self, chunk_idx, get_loss=True):
        smplx_params_chunk = {}
        for key in self.smplx_params.keys():
            if self.is_single(key):
                smplx_params_chunk[key] = self.smplx_params[key]
            else:
                if chunk_idx == -1:
                    smplx_params_chunk[key] = torch.cat(self.smplx_params[key], dim=0)
                else:
                    smplx_params_chunk[key] = self.smplx_params[key][chunk_idx]
        if self.J_regressor_log is not None:
            smplx_params_chunk['J_regressor'] = self.activateJRegressor()

        extra_losses = {}
        if chunk_idx == -1:
            assert not get_loss # assume loss is not needed
            params_list = {key: [] for key in self.optimized_params.keys()}
            for c_idx in range(self.num_chunks):
                for key in self.optimized_params.keys():
                    if self.are_control_points:
                        params_list[key].append(self.optimized_params[key](c_idx, get_loss=False)[0])
                    else:
                        params_list[key].append(self.constraints_activation(self.optimized_params[key][c_idx], key))
            for key in params_list.keys():
                smplx_params_chunk[key] = torch.cat(params_list[key])
        else:
            if self.are_control_points:
                for key in self.optimized_params.keys():
                    smplx_params_chunk[key], loss_key = self.optimized_params[key](chunk_idx, get_loss=get_loss)
                    if len(loss_key) > 0 and get_loss:
                        extra_losses[key] = loss_key
            else:
                for key in self.optimized_params.keys():
                    smplx_params_chunk[key] = self.constraints_activation(self.optimized_params[key][chunk_idx], key)

        # for key in smplx_params_chunk.keys():
        #     assert not torch.isnan(smplx_params_chunk[key]).any(), f'NaNs in SMPLX param {key} for chunk {chunk_idx}!'
        #     assert not torch.isinf(smplx_params_chunk[key]).any(), f'Infs in SMPLX param {key} for chunk {chunk_idx}!'

        return smplx_params_chunk, extra_losses

    def __call__(self, chunk_idx):
        smplx_params_chunk, extra_losses = self.getSMPLXParams(chunk_idx, get_loss=True)
        output = self.model(**smplx_params_chunk)
        return output, extra_losses, smplx_params_chunk


def chunkTensor(tensor, are_control_points, num_chunks, knots_per_chunk, spacing):
    if are_control_points:
        tensor_list = []
        for i in range(num_chunks):
            tensor_list.append(tensor[i*knots_per_chunk*spacing:(i + 1) * knots_per_chunk*spacing])
    else:
        tensor_list = torch.chunk(tensor, num_chunks, dim=0)
    return tensor_list

class SMPLXOptimizer():
    def __init__(self, smplx_kwargs, smplx_params_init, right_hand_constraints, left_hand_constraints, K, RT, mapper, poses_2d=None, poses_2d_weights=None, device='cuda', MAX_ITERS=10000, min_lr=5e-6, chunk_size=100000, loss_fn='huber', optimize_shape=True, optimize_hands=True, optimize_smplx_rt=True, OPTIMIZE_RT=True, OPTIMIZE_RT_FIX_CAM_0=True, MESH_INTERSECTION_LOSS_WEIGHT=0., other_params_init={}, lr=1e-3, SMOOTHNESS_TERMS=[], is_smplfx=True, keys=None, PRINT_EVERY=10, SCHEDULER_PATIENCE=20, GLOBAL_ORIENT_ZERO_X=False, GLOBAL_ORIENT_FIX_Z=True, body_zero_angles_idxs=None, LRs={}, are_control_points=False, REG_NORM_EXPRESSION_W=0.0, BETA_REG_W=1e-3, USE_SIGMOID_ACTIVATION=False, spacing=3, APPLY_SMOOTHNESS_ON_CP=True, apply_constraints_before_decode=False, ANGLES_REG_W=0.0, ANGLES_REG_IDXS=[2, 5, 8, 11, 14, 12, 13], OPTIMIZE_FLAME_SHAPE=True, body_constraints=None, is_single_smplx_rt=False, poses_3d=None, OPTIMIZE_FACES=True, IS_BODY_POSE_6D=False, OPTIMIZE_BODY_POSE=True, depths=None, depths_mask=None, DEPTH_LOSS_WEIGHT=0.0, z_3d=None, Z_3D_LOSS_WEIGHT=0.0, max_dist_contacts=0.075, segmentation_path=None, smplx_model_path=None):
        """
        K : (C, 3, 3) camera intrinsics
        RT : (C, 3, 4) camera extrinsics
        poses_2d: (N, C, J, 2) 2D keypoints
        poses_2d_weights: (N, C, J, 1) weights for 2D keypoints
        poses_3d: (N, J, 3) 3D keypoints for optional 3D supervision
        depths: (N, C, J) relative depths for 2D keypoints
        """
        if poses_2d is not None:
            NUM_POSES = poses_2d.shape[0]
        else:
            assert poses_3d is not None, "Either 2D or 3D poses must be provided!"
            NUM_POSES = poses_3d.shape[0]

        self.mapper = mapper
        for triplet in SMOOTHNESS_TERMS:
            assert triplet[1] in ['velocity', 'acceleration', 'jerk']
            assert isinstance(triplet[2], float)

        self.smoothness_losses = {'velocity': getVelocityStats, 'acceleration': getAccelerationStats, 'jerk': getJerkStats}
        if len(SMOOTHNESS_TERMS) == 0:
            print(PrintColors.WARNING + "Smoothness terms are important for SMPLX optimization!" + PrintColors.ENDC)

        if K is not None:
            assert K.ndim == 3 and K.shape[1:] == (3, 3)
            self.K = K
        if RT is not None:
            assert RT.ndim == 3 and RT.shape[1:] == (3, 4)
            self.P = K @ RT

        self.conv = 'XYZ'

        if OPTIMIZE_RT:
            self.r_extrinsics, self.t_extrinsics = matrix_to_euler_angles(RT[...,:3,:3].detach().clone(), self.conv).requires_grad_(True), RT[...,-1].detach().clone().requires_grad_(True)

        self.input_kwargs = {'right_hand_constraints': right_hand_constraints, 'left_hand_constraints': left_hand_constraints, 'device': device, 'MAX_ITERS': MAX_ITERS, 'min_lr': min_lr, 'chunk_size': chunk_size, 'loss_fn': loss_fn, 'optimize_shape': optimize_shape, 'optimize_hands': optimize_hands, 'K': K, 'RT': RT, 'OPTIMIZE_RT': OPTIMIZE_RT, 'MESH_INTERSECTION_LOSS_WEIGHT': MESH_INTERSECTION_LOSS_WEIGHT, 'lr': lr, 'OPTIMIZE_RT_FIX_CAM_0': OPTIMIZE_RT_FIX_CAM_0, 'SMOOTHNESS_TERMS': SMOOTHNESS_TERMS, 'PRINT_EVERY': PRINT_EVERY, 'SCHEDULER_PATIENCE': SCHEDULER_PATIENCE, 'are_control_points': are_control_points, 'NUM_POSES': NUM_POSES, 'is_smplfx': is_smplfx, 'GLOBAL_ORIENT_ZERO_X': GLOBAL_ORIENT_ZERO_X, 'GLOBAL_ORIENT_FIX_Z': GLOBAL_ORIENT_FIX_Z, 'body_zero_angles_idxs': body_zero_angles_idxs, 'REG_NORM_EXPRESSION_W': REG_NORM_EXPRESSION_W, 'BETA_REG_W': BETA_REG_W, 'USE_SIGMOID_ACTIVATION': USE_SIGMOID_ACTIVATION, 'spacing': spacing, 'APPLY_SMOOTHNESS_ON_CP': APPLY_SMOOTHNESS_ON_CP, 'apply_constraints_before_decode': apply_constraints_before_decode, 'ANGLES_REG_W': ANGLES_REG_W, 'ANGLES_REG_IDXS': ANGLES_REG_IDXS, 'smplx_kwargs': smplx_kwargs, 'OPTIMIZE_FLAME_SHAPE': OPTIMIZE_FLAME_SHAPE, 'body_constraints': body_constraints, 'is_single_smplx_rt': is_single_smplx_rt, 'OPTIMIZE_FACES': OPTIMIZE_FACES, 'optimize_smplx_rt': optimize_smplx_rt, 'IS_BODY_POSE_6D': IS_BODY_POSE_6D, 'OPTIMIZE_BODY_POSE': OPTIMIZE_BODY_POSE, 'DEPTH_LOSS_WEIGHT': DEPTH_LOSS_WEIGHT, 'Z_3D_LOSS_WEIGHT': Z_3D_LOSS_WEIGHT, 'max_dist_contacts': max_dist_contacts}
    
        self.smplx_kwargs = smplx_kwargs

        if LRs is None: LRs = {}
        LRs['lr'] = lr
        self.model = SMPLXWrapper(NUM_POSES, is_smplfx, smplx_kwargs, smplx_params_init, other_params_init, chunk_size, optimize_shape, optimize_hands, optimize_smplx_rt, are_control_points, device, GLOBAL_ORIENT_ZERO_X, GLOBAL_ORIENT_FIX_Z, right_hand_constraints, left_hand_constraints, body_zero_angles_idxs, LRs, USE_SIGMOID_ACTIVATION, spacing, apply_constraints_before_decode, OPTIMIZE_FLAME_SHAPE, body_constraints, is_single_smplx_rt, OPTIMIZE_FACES, IS_BODY_POSE_6D, OPTIMIZE_BODY_POSE)

        if MESH_INTERSECTION_LOSS_WEIGHT > 0:
            self.mesh_intersector = SegmentTriangleMeshIntersectionLossDualHands(torch.LongTensor(self.model.model.faces_original.copy().astype(np.int32)), num_triangles_test=6, target_remove_unlikely_parts=True, far_centroid_thr=0.03, is_intersect_prob=0.5, intersection_tolerance=2e-3, segmentation_path=segmentation_path, smplx_model_path=smplx_model_path).to(device)

        self.num_chunks = self.model.num_chunks
        if poses_2d is not None:
            assert poses_2d.ndim == 4 and poses_2d.shape[-1] == 2 # (N, C, J, 2)
            assert poses_2d_weights.ndim == 4 and poses_2d_weights.shape == poses_2d.shape[:-1] + (1,)
            if poses_2d_weights.max() > 1.0 or poses_2d_weights.min() < 0.0:
                print('WARNING: Weights are not normalized between 0 and 1, normalizing by max value', poses_2d_weights.max().item())

            self.poses_2d_chunks = chunkTensor(poses_2d, are_control_points, self.num_chunks, self.model.knots_per_chunk if are_control_points else None, spacing)
            self.poses_2d_weights_chunks = chunkTensor(poses_2d_weights, are_control_points, self.num_chunks, self.model.knots_per_chunk if are_control_points else None, spacing)
        else:
            self.poses_2d_chunks = None

        if poses_3d is not None:
            assert poses_3d.ndim == 3 and poses_3d.shape[-1] == 3 # (N, J, 3)
            self.poses_3d_chunks = chunkTensor(poses_3d, are_control_points, self.num_chunks, self.model.knots_per_chunk if are_control_points else None, spacing)
        else:
            self.poses_3d_chunks = None

        from .utils_smplx import Contacts
        self.contacts = Contacts()
        if z_3d is not None:
            assert z_3d.ndim == 3 and z_3d.shape[-1] == 3 # (N, J, 3)
            # self.z_3d_chunks = chunkTensor(z_3d, are_control_points, self.num_chunks, self.model.knots_per_chunk if are_control_points else None, spacing)
            z_3d_contacts = self.contacts.get_relative_contacts_joints(z_3d)
            z_3d_weights = getContactWeights(z_3d_contacts, max_dist_contact=self.input_kwargs['max_dist_contacts']).squeeze(-1)
            self.z_3d_chunks = chunkTensor(z_3d_contacts[...,2], are_control_points, self.num_chunks, self.model.knots_per_chunk if are_control_points else None, spacing)
            self.z_3d_weights_chunks = chunkTensor(z_3d_weights, are_control_points, self.num_chunks, self.model.knots_per_chunk if are_control_points else None, spacing)
        else:
            self.z_3d_chunks = None

        if depths is not None:
            assert depths.shape == poses_2d.shape[:-1]
            assert depths_mask.shape == poses_2d.shape[:-1]
            if depths_mask.max() > 1.0 or depths_mask.min() < 0.0:
                print('WARNING: Depth weights are not normalized between 0 and 1, normalizing by max value', depths_mask.max().item())
            
            depths_mask = depths_mask * poses_2d_weights.squeeze(-1).to(depths_mask.device) # apply keypoint weights to depth mask
            self.depth_chunks = chunkTensor(depths, are_control_points, self.num_chunks, self.model.knots_per_chunk if are_control_points else None, spacing)
            self.depth_mask_chunks = chunkTensor(depths_mask, are_control_points, self.num_chunks, self.model.knots_per_chunk if are_control_points else None, spacing)
        else:
            self.depth_chunks, self.depth_mask_chunks = None, None

        self.is_temporal = False
        if keys is not None:
            self.is_temporal = True
            assert len(keys) <= NUM_POSES
            self.are_keys_consecutive = (np.max(keys[1:] - keys[:-1]) == 1).all()
            if are_control_points:
                assert self.are_keys_consecutive, "For control points based optimization, keys must be consecutive!"
                self.keys_chunked_vel, self.keys_chunked_acc, self.keys_chunked_jerk = [], [], []
                for i in range(self.num_chunks):
                    self.keys_chunked_vel.append(getTemporalIndicesBasedKeys(keys[i*self.model.knots_per_chunk*spacing:(i+1)*self.model.knots_per_chunk*spacing], 2))
                    self.keys_chunked_acc.append(getTemporalIndicesBasedKeys(keys[i*self.model.knots_per_chunk*spacing:(i+1)*self.model.knots_per_chunk*spacing], 3))
                    self.keys_chunked_jerk.append(getTemporalIndicesBasedKeys(keys[i*self.model.knots_per_chunk*spacing:(i+1)*self.model.knots_per_chunk*spacing], 4))
            else:
                self.keys_chunked_vel = createChunkedTemporalIndices(keys, NUM_POSES, self.num_chunks, diff=2)
                self.keys_chunked_acc = createChunkedTemporalIndices(keys, NUM_POSES, self.num_chunks, diff=3)
                self.keys_chunked_jerk = createChunkedTemporalIndices(keys, NUM_POSES, self.num_chunks, diff=4)

        self.angles_reg_idxs = []
        for idx in ANGLES_REG_IDXS:
            self.angles_reg_idxs += [idx*3, idx*3+1, idx*3+2]

    def get_optimizer(self):
        params_list = self.model.getOptimizableParamsList()
        if self.input_kwargs['OPTIMIZE_RT']:
            params_list.append({'params': [self.r_extrinsics, self.t_extrinsics], 'lr': self.input_kwargs['lr'] * 0.1, 'name': 'extrinsics'})
        return torch.optim.Adam(params_list)

    def getBestParams(self):
        smplx_params, other_params = self.model.getParams()
        if self.input_kwargs['OPTIMIZE_RT']:
            other_params['RTs'] = torch.cat([euler_angles_to_matrix(self.r_extrinsics, self.conv), self.t_extrinsics.unsqueeze(-1)], dim=-1).detach().clone()
        other_params['input_kwargs'] = self.input_kwargs
        return smplx_params, other_params

    def optimize(self):
        optimizer = self.get_optimizer()
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.8, patience=self.input_kwargs['SCHEDULER_PATIENCE'], min_lr=self.input_kwargs['min_lr'])
        best_loss, best_smplx_params, best_other_params = 1e10, None, None
        loss_fn = getLossFnc(self.input_kwargs['loss_fn'], reduction='none')
        trackers = {}
        if self.poses_2d_chunks is not None:
            trackers = {'body2d': 0.0, 'face2d': 0.0, 'hand2d': 0.0}
            if self.input_kwargs['DEPTH_LOSS_WEIGHT'] > 0.0:
                trackers['depth_loss'] = 0.0
        if self.input_kwargs['Z_3D_LOSS_WEIGHT'] > 0.0:
            trackers['z_3d_loss'] = 0.0
        if self.poses_3d_chunks is not None:
            trackers['pose3d'], trackers['face3d'], trackers['hand3d'] = 0.0, 0.0, 0.0
        if self.input_kwargs['REG_NORM_EXPRESSION_W'] > 0.0:
            trackers['exp_loss'] = 0.0
        if self.input_kwargs['MESH_INTERSECTION_LOSS_WEIGHT'] > 0:
            trackers['int_loss'], trackers['n_ints'] = 0.0, 0.0
        for triplet in self.input_kwargs['SMOOTHNESS_TERMS']:
            trackers[f'{triplet[0]}_{triplet[1]}_loss'] = 0.0
        if self.input_kwargs['ANGLES_REG_W'] > 0.0 and len(self.angles_reg_idxs) > 0:
            trackers['angles_reg_loss'] = 0.0
        time_start = time.time()
        saved_iter = 0
        optimizer_init_lrs = {param_group['name']: param_group['lr'] for param_group in optimizer.param_groups}
        for it in range(self.input_kwargs['MAX_ITERS']):
            for key in trackers.keys():
                trackers[key] = 0.0

            loss_it = 0.0
            for chunk_idx in range(self.num_chunks):
                optimizer.zero_grad()

                output, extra_losses, smplx_params_chunk = self.model(chunk_idx)

                joints = output.joints[:, self.mapper.smplx_idxs]
                loss = 0.0
                # plotMesh([output.vertices[50]], self.model.model.faces, title=f'Iteration {it} chunk {chunk_idx} mesh check', show=True, elev=-90, azim=-90)
                if self.poses_2d_chunks is not None:
                    if self.input_kwargs['OPTIMIZE_RT']:
                        proj_2d, depths = projectP(self.K.to(self.input_kwargs['device']) @ torch.cat([euler_angles_to_matrix(self.r_extrinsics, self.conv), self.t_extrinsics.unsqueeze(-1)], dim=-1).to(self.input_kwargs['device']), joints, with_depth=True)
                    else:
                        proj_2d, depths = projectP(self.P.to(self.input_kwargs['device']), joints, with_depth=True)

                    # if True:
                    if False:
                        from drawer import plotPoses2d
                        idx = 0
                        for cam_v in range(self.poses_2d_chunks[chunk_idx].shape[1]):
                            plotPoses2d([proj_2d[idx, cam_v], self.poses_2d_chunks[chunk_idx][idx, cam_v]], self.mapper.edges, title=f'2D projection check for cam {cam_v}', show=False, legend=['predicted', 'target'], mask=[np.ones(proj_2d.shape[2], dtype=bool), self.poses_2d_weights_chunks[chunk_idx][idx, cam_v].squeeze(-1).bool() > 0.0])
                            # plotPoses2d([torch.cat((proj_2d[idx, cam_v].cpu(), self.poses_2d_chunks[chunk_idx][idx, cam_v]))], getEdgesPairs(proj_2d.shape[-2]), title=f'2D projection check for cam {cam_v}', show=False, legend=['predicted', 'target'])
                            import matplotlib.pyplot as plt
                            plt.show()

                    loss2d = loss_fn(proj_2d, self.poses_2d_chunks[chunk_idx].to(self.input_kwargs['device']))
                    if self.poses_2d_weights_chunks is not None:
                        loss += (loss2d * self.poses_2d_weights_chunks[chunk_idx].to(self.input_kwargs['device'])).sum() / (self.poses_2d_weights_chunks[chunk_idx].to(self.input_kwargs['device']).sum() + 1e-9)
                    else:
                        loss += loss2d.mean()
                    
                    with torch.no_grad():
                        trackers['body2d'] += avgL2distance(proj_2d[:, :, self.mapper.body_subidxs], self.poses_2d_chunks[chunk_idx][:, :, self.mapper.body_subidxs].to(self.input_kwargs['device']), self.poses_2d_weights_chunks[chunk_idx][:, :, self.mapper.body_subidxs].to(self.input_kwargs['device'])).item()
                        trackers['hand2d'] += avgL2distance(proj_2d[:, :, self.mapper.hand_subidxs], self.poses_2d_chunks[chunk_idx][:, :, self.mapper.hand_subidxs].to(self.input_kwargs['device']), self.poses_2d_weights_chunks[chunk_idx][:, :, self.mapper.hand_subidxs].to(self.input_kwargs['device'])).item()
                        trackers['face2d'] += avgL2distance(proj_2d[:, :, self.mapper.face_subidxs], self.poses_2d_chunks[chunk_idx][:, :, self.mapper.face_subidxs].to(self.input_kwargs['device']), self.poses_2d_weights_chunks[chunk_idx][:, :, self.mapper.face_subidxs].to(self.input_kwargs['device'])).item()

                    if self.input_kwargs['DEPTH_LOSS_WEIGHT'] > 0:
                        # depth_loss = compute_ordinal_depth_loss(depths, self.depth_chunks[chunk_idx].to(depths.device), self.depth_mask_chunks[chunk_idx].to(depths.device), global_alignment=self.is_temporal)
                        depth_loss = compute_pearson_depth_loss(depths, self.depth_chunks[chunk_idx].to(depths.device), self.depth_mask_chunks[chunk_idx].to(depths.device))

                        loss += self.input_kwargs['DEPTH_LOSS_WEIGHT'] * depth_loss
                        trackers['depth_loss'] += depth_loss.detach().item()

                if self.poses_3d_chunks is not None:
                    loss3d = loss_fn(joints, self.poses_3d_chunks[chunk_idx].to(self.input_kwargs['device']))
                    if False:
                        from drawer import plotPoses
                        idx = min(100, joints.shape[0]-1)
                        plotPoses([joints[idx]], self.mapper.edges, title='3D check', show=True)
                        plotPoses([joints[idx], self.poses_3d_chunks[chunk_idx][idx]], self.mapper.edges, title='3D supervision check', show=False, legend=['predicted', 'target'])
                        plotPoses([torch.cat((joints[idx].cpu(), self.poses_3d_chunks[chunk_idx][idx]))], getEdgesPairs(len(joints[0])), title='3D supervision check', show=True, legend=['predicted', 'target'])
                    loss += loss3d.mean()
                    with torch.no_grad():
                        trackers['pose3d'] += avgL2distance(joints[:, self.mapper.body_subidxs], self.poses_3d_chunks[chunk_idx][:, self.mapper.body_subidxs].to(self.input_kwargs['device'])).item()
                        trackers['hand3d'] += avgL2distance(joints[:, self.mapper.hand_subidxs], self.poses_3d_chunks[chunk_idx][:, self.mapper.hand_subidxs].to(self.input_kwargs['device'])).item()
                        trackers['face3d'] += avgL2distance(joints[:, self.mapper.face_subidxs], self.poses_3d_chunks[chunk_idx][:, self.mapper.face_subidxs].to(self.input_kwargs['device'])).item()

                if self.input_kwargs['Z_3D_LOSS_WEIGHT'] > 0.0:
                    rel_contacts = self.contacts.get_relative_contacts_joints(output.joints)
                    rel_weights = getContactWeights(rel_contacts, max_dist_contact=self.input_kwargs['max_dist_contacts']).squeeze(-1)
                    rel_weights_combined = rel_weights + self.z_3d_weights_chunks[chunk_idx].to(rel_weights.device) # OR
                    loss_z_3d = ((rel_contacts[...,2] - self.z_3d_chunks[chunk_idx].to(joints.device)).abs() * rel_weights_combined).sum() / (1e-9 + rel_weights_combined.sum())
                    loss += self.input_kwargs['Z_3D_LOSS_WEIGHT'] * loss_z_3d
                    trackers['z_3d_loss'] += loss_z_3d.detach().item()

                for loss_k, loss_v in extra_losses.items():
                    for loss_k_n, loss_v_n in loss_v.items():
                        loss += loss_v_n
                        trackers[f"{loss_k}_{loss_k_n}"] += loss_v_n.detach().item()

                if self.input_kwargs['MESH_INTERSECTION_LOSS_WEIGHT'] > 0:
                    inpenetration_err_, num_total_penetrations = self.mesh_intersector(output, vis=False)#it > 100)
                    loss += self.input_kwargs['MESH_INTERSECTION_LOSS_WEIGHT'] * inpenetration_err_
                    trackers['int_loss'] += (1e3 * inpenetration_err_.detach().item()) # scale 1e3 times
                    trackers['n_ints'] += num_total_penetrations

                if len(self.input_kwargs['SMOOTHNESS_TERMS']) > 0:
                    for triplet in self.input_kwargs['SMOOTHNESS_TERMS']:
                        if self.are_keys_consecutive:
                            if triplet[0] == 'vertices':
                                smooth_loss = self.smoothness_losses[triplet[1]](output.vertices, with_stats=False)[0].mean()
                            else:
                                if self.input_kwargs['are_control_points'] and self.input_kwargs['APPLY_SMOOTHNESS_ON_CP']:
                                    smooth_loss = self.smoothness_losses[triplet[1]](self.model.optimized_params[triplet[0]].chunks[chunk_idx].squeeze(0).T, with_stats=False)[0].mean()
                                else:
                                    smooth_loss = self.smoothness_losses[triplet[1]](smplx_params_chunk[triplet[0]], with_stats=False)[0].mean()
                        else:
                            if triplet[0] == 'vertices':
                                if triplet[1] == 'velocity':
                                    smooth_loss = torch.linalg.norm(output.vertices[self.keys_chunked_vel[chunk_idx][:,1]] - output.vertices[self.keys_chunked_vel[chunk_idx][:,0]], dim=-1).mean()
                                elif triplet[1] == 'acceleration':
                                    smooth_loss = torch.linalg.norm(output.vertices[self.keys_chunked_acc[chunk_idx][:,2]] - 2 * output.vertices[self.keys_chunked_acc[chunk_idx][:,1]] + output.vertices[self.keys_chunked_acc[chunk_idx][:,0]], dim=-1).mean()
                                elif triplet[1] == 'jerk':
                                    smooth_loss = torch.linalg.norm(output.vertices[self.keys_chunked_jerk[chunk_idx][:,3]] - 3 * output.vertices[self.keys_chunked_jerk[chunk_idx][:,2]] + 3 * output.vertices[self.keys_chunked_jerk[chunk_idx][:,1]] - output.vertices[self.keys_chunked_jerk[chunk_idx][:,0]], dim=-1).mean()
                                else:
                                    assert False, "Invalid smoothness type in triplet: " + str(triplet)
                            else:
                                assert False, "THE CODE MUST BE REVISED:"
                                if self.input_kwargs['are_control_points'] and self.input_kwargs['APPLY_SMOOTHNESS_ON_CP']:
                                    smooth_loss = self.smoothness_losses[triplet[1]](self.model.optimized_params[triplet[0]].chunks[chunk_idx].squeeze(0).T, with_stats=False)[0].mean()
                                else:
                                    smooth_loss = self.smoothness_losses[triplet[1]](smplx_params_chunk[triplet[0]][self.keys_chunked[chunk_idx]], with_stats=False)[0].mean()
                        loss += triplet[2] * smooth_loss
                        trackers[f'{triplet[0]}_{triplet[1]}_loss'] += (1e3 * smooth_loss.detach().item()) # scale 1e3 times

                if self.input_kwargs['REG_NORM_EXPRESSION_W'] > 0.0:
                    loss_expression = torch.linalg.norm(smplx_params_chunk['expression'], dim=-1).mean()
                    loss += self.input_kwargs['REG_NORM_EXPRESSION_W'] * loss_expression
                    trackers['exp_loss'] += loss_expression.detach().item()

                if self.input_kwargs['BETA_REG_W'] > 0.0:
                    if self.input_kwargs['optimize_shape']:
                        loss += self.input_kwargs['BETA_REG_W'] * torch.linalg.norm(smplx_params_chunk['betas'], dim=-1).mean()
                    if self.input_kwargs['OPTIMIZE_FLAME_SHAPE'] and self.input_kwargs['is_smplfx']:
                        loss += self.input_kwargs['BETA_REG_W'] * torch.linalg.norm(smplx_params_chunk['shape'], dim=-1).mean()

                if self.input_kwargs['ANGLES_REG_W'] > 0.0 and len(self.angles_reg_idxs) > 0:
                    loss_angles_reg = torch.abs(smplx_params_chunk['body_pose'][:, self.angles_reg_idxs]).mean()
                    loss += self.input_kwargs['ANGLES_REG_W'] * loss_angles_reg
                    trackers['angles_reg_loss'] += loss_angles_reg.detach().item()

                loss.backward()

                if self.input_kwargs['OPTIMIZE_RT'] and self.input_kwargs.get('opt_rt_extrinsics_fix_first_view', False):
                    self.r_extrinsics.grad[0].zero_()
                    self.t_extrinsics.grad[0].zero_()

                optimizer.step()
                scheduler.step(loss)
                for optimizer_param_group in optimizer.param_groups:
                    optimizer_param_group['lr'] = min(optimizer_param_group['lr'], optimizer_init_lrs[optimizer_param_group['name']])

                loss_it += loss.detach().item()

                self.model.apply_constraints(chunk_idx)
    
            if it % self.input_kwargs['PRINT_EVERY'] == 0:
                log_str = f"Iter {it}, Loss {loss_it:.3f}, time {(time.time() - time_start)/60:.2f}m, lr {optimizer.param_groups[0]['lr']:.2e}"
                for key in trackers.keys():
                    log_str += f", {key} {trackers[key]/(chunk_idx+1):.3f}"
                if self.input_kwargs['optimize_shape']:
                    log_str += f' beta {torch.linalg.norm(self.model.smplx_params["betas"]).item():.3f}'
                if self.input_kwargs['is_smplfx'] and self.input_kwargs['OPTIMIZE_FLAME_SHAPE']:
                    log_str += f' shape {torch.linalg.norm(self.model.smplx_params["shape"]).item():.3f}'
                print(log_str)

            if best_loss > loss_it and it > 5: # never save on the first few iterations, as the loss can be very unstable at the beginning
                best_loss = loss_it
                saved_iter = it
                # print('SAVE BEST PARAMS at iter', it, 'with loss', best_loss)
                best_smplx_params, best_other_params = self.getBestParams()

            if optimizer.param_groups[0]['lr'] <= self.input_kwargs['min_lr']:
                print(f"Stopping optimization at iteration {it} due to learning rate reaching minimum.")
                break

        # torch.save({'smplx': best_smplx_params, 'others': best_other_params}, '')
        return best_smplx_params, best_other_params



"""Pure skeleton drawing functions using OpenCV.

All functions draw directly onto a BGR frame (numpy array) in-place.
Colors are in BGR format for OpenCV.
"""

from __future__ import annotations

import numpy as np

_INT32_MAX = 2**31 - 1
_INT32_MIN = -(2**31)

# ============================================================================
# Connection Constants
# ============================================================================

# MediaPipe Pose 33-landmark body connections
MEDIAPIPE_BODY_CONNECTIONS: list[tuple[int, int]] = [
    (0, 1),
    (1, 2),
    (2, 3),
    (3, 7),  # Right eye
    (0, 4),
    (4, 5),
    (5, 6),
    (6, 8),  # Left eye
    (9, 10),  # Mouth
    (11, 12),  # Shoulders
    (11, 13),
    (13, 15),  # Left arm
    (12, 14),
    (14, 16),  # Right arm
    (15, 17),
    (15, 19),
    (15, 21),
    (17, 19),  # Left wrist
    (16, 18),
    (16, 20),
    (16, 22),
    (18, 20),  # Right wrist
    (11, 23),
    (12, 24),
    (23, 24),  # Torso
    (23, 25),
    (25, 27),
    (27, 29),
    (27, 31),
    (29, 31),  # Left leg
    (24, 26),
    (26, 28),
    (28, 30),
    (28, 32),
    (30, 32),  # Right leg
]

# 21-joint hand connections (shared by MediaPipe hands and WiLoR)
HAND_CONNECTIONS: list[tuple[int, int]] = [
    (0, 1),
    (1, 2),
    (2, 3),
    (3, 4),  # Thumb
    (0, 5),
    (5, 6),
    (6, 7),
    (7, 8),  # Index
    (0, 9),
    (9, 10),
    (10, 11),
    (11, 12),  # Middle
    (0, 13),
    (13, 14),
    (14, 15),
    (15, 16),  # Ring
    (0, 17),
    (17, 18),
    (18, 19),
    (19, 20),  # Pinky
    (5, 9),
    (9, 13),
    (13, 17),  # Palm
]

# COCO 17-body-joint connections
# Joint order: 0=nose, 1=L_eye, 2=R_eye, 3=L_ear, 4=R_ear,
#   5=L_shoulder, 6=R_shoulder, 7=L_elbow, 8=R_elbow,
#   9=L_wrist, 10=R_wrist, 11=L_hip, 12=R_hip,
#   13=L_knee, 14=R_knee, 15=L_ankle, 16=R_ankle
# COCO upper-body connections only (no legs 13-16, no feet 17-22)
COCO_UPPER_BODY_CONNECTIONS: list[tuple[int, int]] = [
    (0, 1),
    (0, 2),
    (1, 3),
    (2, 4),  # Head
    (5, 6),  # Shoulders
    (5, 7),
    (7, 9),  # Left arm
    (6, 8),
    (8, 10),  # Right arm
    (5, 11),
    (6, 12),
    (11, 12),  # Torso
]

# Full COCO 17-body connections (upper body + legs)
COCO_BODY_CONNECTIONS: list[tuple[int, int]] = COCO_UPPER_BODY_CONNECTIONS + [
    (11, 13),
    (13, 15),  # Left leg
    (12, 14),
    (14, 16),  # Right leg
]

# SMPL-X 22-body-joint connections
# Joint order: 0=pelvis, 1=L_hip, 2=R_hip, 3=spine1, 4=L_knee, 5=R_knee,
#   6=spine2, 7=L_ankle, 8=R_ankle, 9=spine3, 10=L_foot, 11=R_foot,
#   12=neck, 13=L_collar, 14=R_collar, 15=head,
#   16=L_shoulder, 17=R_shoulder, 18=L_elbow, 19=R_elbow,
#   20=L_wrist, 21=R_wrist
SMPLX_BODY_CONNECTIONS: list[tuple[int, int]] = [
    # Spine: pelvis -> spine1 -> spine2 -> spine3 -> neck -> head
    (0, 3),
    (3, 6),
    (6, 9),
    (9, 12),
    (12, 15),
    # Hip bar: pelvis -> hips
    (0, 1),
    (0, 2),
    # Left arm: spine3 -> L_collar -> L_shoulder -> L_elbow -> L_wrist
    (9, 13),
    (13, 16),
    (16, 18),
    (18, 20),
    # Right arm: spine3 -> R_collar -> R_shoulder -> R_elbow -> R_wrist
    (9, 14),
    (14, 17),
    (17, 19),
    (19, 21),
    # Left leg: L_hip -> L_knee -> L_ankle -> L_foot
    (1, 4),
    (4, 7),
    (7, 10),
    # Right leg: R_hip -> R_knee -> R_ankle -> R_foot
    (2, 5),
    (5, 8),
    (8, 11),
]

# SMPL-X hand joints: 15 per hand, NO wrist (wrist is body joint 20/21).
# Order: index(0-2), middle(3-5), pinky(6-8), ring(9-11), thumb(12-14)
# Each triplet is MCP -> PIP -> DIP.  Intra-finger chains only;
# wrist-to-MCP connections are drawn separately using body wrist joint.
SMPLX_HAND_CONNECTIONS: list[tuple[int, int]] = [
    (0, 1),
    (1, 2),  # Index: MCP -> PIP -> DIP
    (3, 4),
    (4, 5),  # Middle: MCP -> PIP -> DIP
    (6, 7),
    (7, 8),  # Pinky: MCP -> PIP -> DIP
    (9, 10),
    (10, 11),  # Ring: MCP -> PIP -> DIP
    (12, 13),
    (13, 14),  # Thumb: CMC -> IP -> TIP
]

# MCP (base) joint local indices for connecting to wrist
SMPLX_HAND_MCP_INDICES: list[int] = [0, 3, 6, 9, 12]  # index, middle, pinky, ring, thumb

# ============================================================================
# Color Scheme (BGR for OpenCV)
# ============================================================================

COLOR_BODY = (250, 165, 96)  # Blue-ish (matches PoseViewer2D #60a5fa)
COLOR_LEFT_HAND = (74, 222, 128)  # Green (matches #4ade80)
COLOR_RIGHT_HAND = (168, 85, 247)  # Purple (matches #f755a8 -> BGR)
COLOR_JOINT = (255, 255, 255)  # White joint dots
COLOR_FACE = (0, 200, 255)  # Orange/amber for face box
COLOR_FACE_LANDMARK = (0, 220, 255)  # Slightly brighter amber for landmarks


def _line_thickness(h: int, w: int) -> int:
    """Scale line thickness with frame resolution."""
    return max(1, int(min(h, w) / 300))


def _circle_radius(h: int, w: int) -> int:
    """Scale joint circle radius with frame resolution."""
    return max(1, int(min(h, w) / 200))


def _draw_connections(
    frame: np.ndarray,
    keypoints: np.ndarray,
    connections: list[tuple[int, int]],
    color: tuple[int, int, int],
    thickness: int,
    confidence: np.ndarray | None = None,
    min_confidence: float = 0.1,
) -> None:
    """Draw skeleton connections on a frame."""
    import cv2

    n_kpts = len(keypoints)
    for i, j in connections:
        if i >= n_kpts or j >= n_kpts:
            continue

        # Skip non-finite coordinates (NaN, Inf)
        if not (np.isfinite(keypoints[i][0]) and np.isfinite(keypoints[i][1])):
            continue
        if not (np.isfinite(keypoints[j][0]) and np.isfinite(keypoints[j][1])):
            continue

        # Clamp to int32 range for OpenCV compatibility
        x1 = max(_INT32_MIN, min(_INT32_MAX, int(keypoints[i][0])))
        y1 = max(_INT32_MIN, min(_INT32_MAX, int(keypoints[i][1])))
        x2 = max(_INT32_MIN, min(_INT32_MAX, int(keypoints[j][0])))
        y2 = max(_INT32_MIN, min(_INT32_MAX, int(keypoints[j][1])))

        # Skip zero-coordinate points (no detection)
        if (x1 == 0 and y1 == 0) or (x2 == 0 and y2 == 0):
            continue

        # Skip low-confidence joints
        if confidence is not None:
            if i < len(confidence) and confidence[i] < min_confidence:
                continue
            if j < len(confidence) and confidence[j] < min_confidence:
                continue

        cv2.line(frame, (x1, y1), (x2, y2), color, thickness, cv2.LINE_AA)


def _draw_joints(
    frame: np.ndarray,
    keypoints: np.ndarray,
    radius: int,
    color: tuple[int, int, int] = COLOR_JOINT,
    confidence: np.ndarray | None = None,
    min_confidence: float = 0.1,
) -> None:
    """Draw joint circles on a frame."""
    import cv2

    for idx, kpt in enumerate(keypoints):
        if not (np.isfinite(kpt[0]) and np.isfinite(kpt[1])):
            continue
        x = max(_INT32_MIN, min(_INT32_MAX, int(kpt[0])))
        y = max(_INT32_MIN, min(_INT32_MAX, int(kpt[1])))
        if x == 0 and y == 0:
            continue
        if confidence is not None and idx < len(confidence) and confidence[idx] < min_confidence:
            continue
        cv2.circle(frame, (x, y), radius, color, -1, cv2.LINE_AA)


# ============================================================================
# Public Drawing Functions
# ============================================================================


def draw_mediapipe_skeleton(
    frame: np.ndarray,
    keypoints: np.ndarray,
    confidence: np.ndarray | None = None,
    hands_l: np.ndarray | None = None,
    hands_r: np.ndarray | None = None,
) -> None:
    """Draw MediaPipe skeleton on a video frame.

    Args:
        frame: BGR video frame (H, W, 3), modified in-place.
        keypoints: Body landmarks shape (33, 2|3). If values < 2, treated as
            normalized 0-1 and scaled to frame dimensions.
        confidence: Per-joint confidence (33,). Joints below 0.1 are hidden.
        hands_l: Left hand landmarks (21, 2|3). Same normalization rules.
        hands_r: Right hand landmarks (21, 2|3). Same normalization rules.
    """
    h, w = frame.shape[:2]
    thickness = _line_thickness(h, w)
    radius = _circle_radius(h, w)

    def to_pixels(kpts: np.ndarray) -> np.ndarray:
        """Convert normalized coords to pixel coords if needed."""
        kpts = np.asarray(kpts, dtype=np.float64)
        result = kpts[:, :2].copy()
        # Use nanmax to handle NaN values; check if all finite values are in [0, 1]
        finite_mask = np.isfinite(result)
        if finite_mask.any():
            max_val = np.nanmax(np.abs(result[finite_mask]))
            if max_val <= 1.0:
                result[:, 0] *= w
                result[:, 1] *= h
        return result

    body_px = to_pixels(keypoints)

    # Draw body
    _draw_connections(frame, body_px, MEDIAPIPE_BODY_CONNECTIONS, COLOR_BODY, thickness, confidence)
    _draw_joints(frame, body_px, radius, COLOR_BODY, confidence)

    # Draw left hand
    if hands_l is not None and len(hands_l) > 0:
        lh_px = to_pixels(hands_l)
        _draw_connections(frame, lh_px, HAND_CONNECTIONS, COLOR_LEFT_HAND, thickness)
        _draw_joints(frame, lh_px, radius, COLOR_LEFT_HAND)

    # Draw right hand
    if hands_r is not None and len(hands_r) > 0:
        rh_px = to_pixels(hands_r)
        _draw_connections(frame, rh_px, HAND_CONNECTIONS, COLOR_RIGHT_HAND, thickness)
        _draw_joints(frame, rh_px, radius, COLOR_RIGHT_HAND)


def draw_wilor_hands(
    frame: np.ndarray,
    kpts_2d: np.ndarray,
    right_flags: np.ndarray | None = None,
) -> None:
    """Draw WiLoR hand detections on a video frame.

    Args:
        frame: BGR video frame (H, W, 3), modified in-place.
        kpts_2d: 2D keypoints shape (N, 21, 2) in pixel coords.
        right_flags: Boolean array (N,) -- True means right hand.
            If None, alternates left/right.
    """
    h, w = frame.shape[:2]
    thickness = _line_thickness(h, w)
    radius = _circle_radius(h, w)

    for det_idx in range(len(kpts_2d)):
        hand_kpts = kpts_2d[det_idx]  # (21, 2)
        is_right = right_flags[det_idx] if right_flags is not None else (det_idx % 2 == 1)
        color = COLOR_RIGHT_HAND if is_right else COLOR_LEFT_HAND

        _draw_connections(frame, hand_kpts, HAND_CONNECTIONS, color, thickness)
        _draw_joints(frame, hand_kpts, radius, color)


def draw_nlf_body(
    frame: np.ndarray,
    joints2d: np.ndarray,
) -> None:
    """Draw NLF/SMPL-X body skeleton on a video frame.

    Args:
        frame: BGR video frame (H, W, 3), modified in-place.
        joints2d: 2D joint positions shape (N, 55, 2) in pixel coords.
            For each detection: 22 body + 15 left hand + 15 right hand + 3 extra.
    """
    h, w = frame.shape[:2]
    thickness = _line_thickness(h, w)
    radius = _circle_radius(h, w)

    for det_idx in range(len(joints2d)):
        joints = joints2d[det_idx]  # (55, 2)

        # SMPL-X 55-joint layout:
        #   0-21:  22 body joints (20=L_wrist, 21=R_wrist)
        #   22:    jaw
        #   23:    left eye
        #   24:    right eye
        #   25-39: 15 left hand joints  (index, middle, pinky, ring, thumb × 3)
        #   40-54: 15 right hand joints (same order)

        # Body joints (0-21)
        body_joints = joints[:22]
        _draw_connections(frame, body_joints, SMPLX_BODY_CONNECTIONS, COLOR_BODY, thickness)
        _draw_joints(frame, body_joints, radius, COLOR_BODY)

        # Jaw / eyes (22-24) -- draw as dots only
        extra_joints = joints[22:25]
        if len(extra_joints) > 0:
            _draw_joints(frame, extra_joints, radius, COLOR_BODY)

        # Left hand (25-39, mapped to 0-14 for SMPLX_HAND_CONNECTIONS)
        left_hand = joints[25:40]
        if len(left_hand) == 15:
            _draw_connections(frame, left_hand, SMPLX_HAND_CONNECTIONS, COLOR_LEFT_HAND, thickness)
            _draw_joints(frame, left_hand, radius, COLOR_LEFT_HAND)
            # Connect wrist (body joint 20) to each finger's MCP
            l_wrist = joints[20:21]  # (1, 2)
            for mcp_idx in SMPLX_HAND_MCP_INDICES:
                mcp = left_hand[mcp_idx : mcp_idx + 1]  # (1, 2)
                _draw_connections(
                    frame,
                    np.concatenate([l_wrist, mcp], axis=0),
                    [(0, 1)],
                    COLOR_LEFT_HAND,
                    thickness,
                )

        # Right hand (40-54, mapped to 0-14)
        right_hand = joints[40:55]
        if len(right_hand) == 15:
            _draw_connections(
                frame,
                right_hand,
                SMPLX_HAND_CONNECTIONS,
                COLOR_RIGHT_HAND,
                thickness,
            )
            _draw_joints(frame, right_hand, radius, COLOR_RIGHT_HAND)
            # Connect wrist (body joint 21) to each finger's MCP
            r_wrist = joints[21:22]  # (1, 2)
            for mcp_idx in SMPLX_HAND_MCP_INDICES:
                mcp = right_hand[mcp_idx : mcp_idx + 1]  # (1, 2)
                _draw_connections(
                    frame,
                    np.concatenate([r_wrist, mcp], axis=0),
                    [(0, 1)],
                    COLOR_RIGHT_HAND,
                    thickness,
                )


def draw_teaser_face(
    frame: np.ndarray,
    boxes: np.ndarray,
    scores: np.ndarray | None = None,
    jaw_pose: np.ndarray | None = None,
    expression: np.ndarray | None = None,
    eyelid_params: np.ndarray | None = None,
) -> None:
    """Draw TEASER face detection boxes and FLAME parameter HUD.

    Args:
        frame: BGR video frame (H, W, 3), modified in-place.
        boxes: Bounding boxes (N, 4) in xyxy pixel coords.
        scores: Detection confidence (N,).
        jaw_pose: FLAME jaw pose (N, 3).
        expression: FLAME expression coefficients (N, 50).
        eyelid_params: FLAME eyelid parameters (N, 2).
    """
    import cv2

    h, w = frame.shape[:2]
    thickness = _line_thickness(h, w)
    font_scale = max(0.3, min(h, w) / 800)
    font = cv2.FONT_HERSHEY_SIMPLEX

    for i in range(len(boxes)):
        box = boxes[i]
        if not np.isfinite(box).all():
            continue

        x1, y1, x2, y2 = int(box[0]), int(box[1]), int(box[2]), int(box[3])

        # Draw bounding box
        cv2.rectangle(frame, (x1, y1), (x2, y2), COLOR_FACE, thickness)

        # Build text lines for HUD
        lines: list[str] = []
        if scores is not None and i < len(scores):
            lines.append(f"conf: {float(scores[i]):.2f}")
        if jaw_pose is not None and i < len(jaw_pose):
            jp = jaw_pose[i]
            lines.append(f"jaw: {jp[0]:.2f} {jp[1]:.2f} {jp[2]:.2f}")
        if eyelid_params is not None and i < len(eyelid_params):
            ep = eyelid_params[i]
            lines.append(f"eyelid: {ep[0]:.2f} {ep[1]:.2f}")
        if expression is not None and i < len(expression):
            # Show L2 norm of expression vector as a compact summary
            expr_norm = float(np.linalg.norm(expression[i]))
            lines.append(f"expr|: {expr_norm:.1f}")

        # Draw text above the box
        text_y = max(y1 - 4, 12)
        for line in reversed(lines):
            (tw, th), _ = cv2.getTextSize(line, font, font_scale, 1)
            # Background rectangle for readability
            cv2.rectangle(
                frame,
                (x1, text_y - th - 4),
                (x1 + tw + 4, text_y + 2),
                (0, 0, 0),
                -1,
            )
            cv2.putText(frame, line, (x1 + 2, text_y), font, font_scale, COLOR_FACE, 1, cv2.LINE_AA)
            text_y -= th + 6


# 68-point FAN landmark connections (face contour, brows, nose, eyes, mouth)
FLAME_FACE_CONNECTIONS: list[tuple[int, int]] = [
    # Jaw contour (0-16)
    *[(i, i + 1) for i in range(16)],
    # Right eyebrow (17-21)
    *[(i, i + 1) for i in range(17, 21)],
    # Left eyebrow (22-26)
    *[(i, i + 1) for i in range(22, 26)],
    # Nose bridge (27-30)
    *[(i, i + 1) for i in range(27, 30)],
    # Nose bottom (31-35)
    *[(i, i + 1) for i in range(31, 35)],
    # Right eye (36-41, closed loop)
    *[(i, i + 1) for i in range(36, 41)],
    (41, 36),
    # Left eye (42-47, closed loop)
    *[(i, i + 1) for i in range(42, 47)],
    (47, 42),
    # Outer mouth (48-59, closed loop)
    *[(i, i + 1) for i in range(48, 59)],
    (59, 48),
    # Inner mouth (60-67, closed loop)
    *[(i, i + 1) for i in range(60, 67)],
    (67, 60),
]


def draw_teaser_landmarks(
    frame: np.ndarray,
    landmarks_2d: np.ndarray,
) -> None:
    """Draw 68-point FAN face landmarks from FLAME on a video frame.

    Args:
        frame: BGR video frame (H, W, 3), modified in-place.
        landmarks_2d: 2D landmarks shape (N, 68, 2) in pixel coords.
            N is the number of face detections for this frame.
    """
    h, w = frame.shape[:2]
    thickness = _line_thickness(h, w)
    radius = max(1, _circle_radius(h, w) // 2)  # Smaller dots for face

    for det_idx in range(len(landmarks_2d)):
        lmk = landmarks_2d[det_idx]  # (68, 2)
        _draw_connections(frame, lmk, FLAME_FACE_CONNECTIONS, COLOR_FACE_LANDMARK, thickness)
        _draw_joints(frame, lmk, radius, COLOR_FACE_LANDMARK)


def draw_rtmpose_wholebody(
    frame: np.ndarray,
    keypoints: np.ndarray,
) -> None:
    """Draw COCO-WholeBody skeleton on a video frame.

    Args:
        frame: BGR video frame (H, W, 3), modified in-place.
        keypoints: Whole-body landmarks shape ``(N, 133, 3)`` in pixel coords.
            Layout: body(0-16) + foot(17-22) + face(23-90) + left_hand(91-111)
            + right_hand(112-132).  Channel 2 is confidence.
    """
    h, w = frame.shape[:2]
    thickness = _line_thickness(h, w)
    radius = _circle_radius(h, w)
    face_radius = max(1, radius // 2)

    for det_idx in range(len(keypoints)):
        kpts = keypoints[det_idx]  # (133, 3)
        conf = kpts[:, 2]

        # Upper body only (0-12): head, arms, torso — skip legs (13-16) and feet (17-22)
        body_xy = kpts[:17, :2]
        _draw_connections(
            frame,
            body_xy,
            COCO_UPPER_BODY_CONNECTIONS,
            COLOR_BODY,
            thickness,
            conf[:17],
        )
        _draw_joints(frame, body_xy[:13], radius, COLOR_BODY, conf[:13])

        # Face (23-90, 68 points) — use FLAME connections (same 68-point layout)
        face_xy = kpts[23:91, :2]
        _draw_connections(
            frame,
            face_xy,
            FLAME_FACE_CONNECTIONS,
            COLOR_FACE_LANDMARK,
            thickness,
        )
        _draw_joints(frame, face_xy, face_radius, COLOR_FACE_LANDMARK)

        # Left hand (91-111, 21 points)
        left_hand = kpts[91:112, :2]
        _draw_connections(
            frame,
            left_hand,
            HAND_CONNECTIONS,
            COLOR_LEFT_HAND,
            thickness,
        )
        _draw_joints(frame, left_hand, radius, COLOR_LEFT_HAND)

        # Right hand (112-132, 21 points)
        right_hand = kpts[112:133, :2]
        _draw_connections(
            frame,
            right_hand,
            HAND_CONNECTIONS,
            COLOR_RIGHT_HAND,
            thickness,
        )
        _draw_joints(frame, right_hand, radius, COLOR_RIGHT_HAND)



"""
Tests for the embedding visualization API endpoints.

Tests cover:
- Thread-safe SignRep initialization
- Embedding cache operations
- API endpoint responses
- Dimensionality reduction (PCA, t-SNE, UMAP)
- Clustering (kmeans, hdbscan)
- Error handling (missing dataset, invalid params, SignRep unavailable)
- Rate limiting (if slowapi is available)
"""

from __future__ import annotations

import threading
import time
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

# Skip all tests if fastapi not installed
pytest.importorskip("fastapi")

# Mark all tests in this module as API tests
pytestmark = pytest.mark.api

from fastapi import HTTPException
from fastapi.testclient import TestClient

from sltk.api import dependencies as api_deps
from sltk.api.main import app

# ============================================================================
# Fixtures
# ============================================================================


@pytest.fixture
def client():
    """Create a test client for the FastAPI app."""
    return TestClient(app)


@pytest.fixture
def mock_embeddings():
    """Create mock embeddings for testing."""
    np.random.seed(42)
    return np.random.randn(50, 768).astype(np.float32)


@pytest.fixture
def mock_metadata():
    """Create mock metadata matching the mock embeddings."""
    return {
        "sample_ids": [f"sample_{i:03d}" for i in range(50)],
        "glosses": [f"GLOSS_{i % 10}" for i in range(50)],
        "signer_ids": [f"signer_{i % 5}" for i in range(50)],
        "video_paths": [f"/path/to/video_{i:03d}.mp4" for i in range(50)],
    }


@pytest.fixture
def mock_signrep_embedder():
    """Create a mock SignRep embedder."""
    embedder = MagicMock()
    embedder.is_loaded.return_value = True
    embedder.embedding_dim = 768
    embedder.config = MagicMock()
    embedder.config.checkpoint_path = "/mock/path/to/ckpt.pt"

    def mock_embed_video(video_path, aggregation="mean"):
        np.random.seed(hash(str(video_path)) % (2**32))
        return np.random.randn(768).astype(np.float32)

    embedder.embed_video.side_effect = mock_embed_video
    return embedder


@pytest.fixture
def mock_dataset():
    """Create a mock dataset for testing."""
    dataset = MagicMock()
    dataset._index = [f"sample_{i:03d}" for i in range(100)]

    def mock_load_sample(sample_id):
        sample = MagicMock()
        sample.id = sample_id
        sample.video_path = f"/path/to/{sample_id}.mp4"
        sample.glosses = [f"GLOSS_{hash(sample_id) % 10}"]
        sample.signer_id = f"signer_{hash(sample_id) % 5}"
        return sample

    dataset._load_sample.side_effect = mock_load_sample
    return dataset


@pytest.fixture
def patched_cache(mock_embeddings, mock_metadata):
    """Patch the embedding cache with test data."""
    from sltk.api.routers import embeddings as emb_router

    # Save original cache state
    original_cache = emb_router._embedding_cache.copy()

    # Add test data to cache
    emb_router._embedding_cache["test_dataset"] = (
        mock_embeddings,
        mock_metadata,
        time.time(),
    )

    yield

    # Restore original cache
    emb_router._embedding_cache.clear()
    emb_router._embedding_cache.update(original_cache)


@pytest.fixture
def clear_signrep_state():
    """Clear SignRep embedder state before and after test."""
    from sltk.api.routers import embeddings as emb_router

    # Save original state
    original_embedder = emb_router._signrep_embedder
    original_error = emb_router._signrep_init_error

    # Clear state
    emb_router._signrep_embedder = None
    emb_router._signrep_init_error = None

    yield

    # Restore original state
    emb_router._signrep_embedder = original_embedder
    emb_router._signrep_init_error = original_error


@pytest.fixture
def force_signrep_unavailable():
    """Force SignRep to be unavailable for testing error handling."""
    from sltk.api.routers import embeddings as emb_router

    # Save original state
    original_embedder = emb_router._signrep_embedder
    original_error = emb_router._signrep_init_error

    # Force unavailable state
    emb_router._signrep_embedder = None
    emb_router._signrep_init_error = "SignRep not installed (test mock)"

    yield

    # Restore original state
    emb_router._signrep_embedder = original_embedder
    emb_router._signrep_init_error = original_error


@pytest.fixture
def clear_embedding_cache():
    """Clear embedding cache before and after test."""
    from sltk.api.routers import embeddings as emb_router

    original_cache = emb_router._embedding_cache.copy()
    emb_router._embedding_cache.clear()

    yield

    emb_router._embedding_cache.clear()
    emb_router._embedding_cache.update(original_cache)


# ============================================================================
# Thread-safe SignRep initialization tests
# ============================================================================


class TestSignRepInitialization:
    """Tests for thread-safe SignRep initialization."""

    def test_signrep_lazy_initialization_with_forced_error(self, clear_signrep_state):
        """Test that SignRep is lazily initialized on first access and errors are handled."""
        from sltk.api.routers import embeddings as emb_router

        # Clear state
        assert emb_router._signrep_embedder is None

        # Force an import error by patching the import
        with patch(
            "sltk.embedding.signrep.SignRepEmbedder",
            side_effect=ImportError("SignRep not installed (test)"),
        ):
            # Also need to clear cached modules to force re-import
            import sys

            # Remove cached import if exists
            emb_router._signrep_init_error = None

            # Attempting to get embedder should try to initialize and fail
            with pytest.raises(HTTPException) as exc_info:
                emb_router._get_signrep_embedder()

            # The actual error may vary depending on environment, just verify 503
            assert exc_info.value.status_code == 503

    def test_signrep_init_error_cached_behavior(self, clear_signrep_state):
        """Test that initialization errors are cached and reused."""
        from sltk.api.routers import embeddings as emb_router

        # Manually set an error to test caching behavior
        emb_router._signrep_init_error = "Test error message"

        # Should immediately raise without attempting initialization
        with pytest.raises(HTTPException) as exc_info:
            emb_router._get_signrep_embedder()

        assert exc_info.value.status_code == 503
        assert "unavailable" in exc_info.value.detail.lower()
        assert "Test error message" in exc_info.value.detail

    def test_signrep_double_check_locking_pattern(self, clear_signrep_state, mock_signrep_embedder):
        """Test the double-check locking pattern for thread safety."""
        from sltk.api.routers import embeddings as emb_router

        # Set up the embedder directly to test caching
        emb_router._signrep_embedder = mock_signrep_embedder

        # Fast path should return cached embedder
        result = emb_router._get_signrep_embedder()
        assert result is mock_signrep_embedder


class TestSignRepMocked:
    """Tests with mocked SignRep embedder."""

    def test_get_signrep_embedder_cached(self, clear_signrep_state, mock_signrep_embedder):
        """Test that embedder is cached after first access."""
        from sltk.api.routers import embeddings as emb_router

        # Set up cache directly
        emb_router._signrep_embedder = mock_signrep_embedder

        # Should return cached instance without trying to import
        embedder = emb_router._get_signrep_embedder()
        assert embedder is mock_signrep_embedder

    def test_get_signrep_embedder_properties(self, clear_signrep_state, mock_signrep_embedder):
        """Test that cached embedder has expected properties."""
        from sltk.api.routers import embeddings as emb_router

        emb_router._signrep_embedder = mock_signrep_embedder

        embedder = emb_router._get_signrep_embedder()
        assert embedder.is_loaded() is True
        assert embedder.embedding_dim == 768


# ============================================================================
# Embedding cache tests
# ============================================================================


class TestEmbeddingCache:
    """Tests for embedding cache operations."""

    def test_cache_embeddings(self, clear_embedding_cache, mock_embeddings, mock_metadata):
        """Test caching embeddings."""
        from sltk.api.routers import embeddings as emb_router

        emb_router._cache_embeddings("test_ds", mock_embeddings, mock_metadata)

        # Verify cache
        assert "test_ds" in emb_router._embedding_cache
        cached_emb, cached_meta, cached_time = emb_router._embedding_cache["test_ds"]
        np.testing.assert_array_equal(cached_emb, mock_embeddings)
        assert cached_meta == mock_metadata

    def test_get_cached_embeddings(self, clear_embedding_cache, mock_embeddings, mock_metadata):
        """Test retrieving cached embeddings."""
        from sltk.api.routers import embeddings as emb_router

        # Cache embeddings
        emb_router._cache_embeddings("test_ds", mock_embeddings, mock_metadata)

        # Retrieve
        result = emb_router._get_cached_embeddings("test_ds")
        assert result is not None
        emb, meta = result
        np.testing.assert_array_equal(emb, mock_embeddings)

    def test_get_cached_embeddings_expired(
        self, clear_embedding_cache, mock_embeddings, mock_metadata
    ):
        """Test that expired cache entries are removed."""
        from sltk.api.routers import embeddings as emb_router

        # Cache with expired timestamp
        expired_time = time.time() - emb_router.EMBEDDING_CACHE_TTL - 100
        emb_router._embedding_cache["expired_ds"] = (
            mock_embeddings,
            mock_metadata,
            expired_time,
        )

        # Should return None for expired entry
        result = emb_router._get_cached_embeddings("expired_ds")
        assert result is None

        # Entry should be removed from cache
        assert "expired_ds" not in emb_router._embedding_cache

    def test_get_cached_embeddings_not_found(self, clear_embedding_cache):
        """Test retrieving non-existent cache entry."""
        from sltk.api.routers import embeddings as emb_router

        result = emb_router._get_cached_embeddings("nonexistent")
        assert result is None

    def test_cache_thread_safety(self, clear_embedding_cache, mock_embeddings, mock_metadata):
        """Test that cache operations are thread-safe."""
        from sltk.api.routers import embeddings as emb_router

        results = []

        def cache_and_get(dataset_name):
            emb_router._cache_embeddings(dataset_name, mock_embeddings, mock_metadata)
            result = emb_router._get_cached_embeddings(dataset_name)
            results.append(result is not None)

        threads = [threading.Thread(target=cache_and_get, args=(f"ds_{i}",)) for i in range(10)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert all(results)


# ============================================================================
# API endpoint tests
# ============================================================================


class TestEmbeddingStatusEndpoint:
    """Tests for GET /api/embeddings/status/{dataset}."""

    def test_embedding_status_not_cached(
        self, client, clear_embedding_cache, force_signrep_unavailable
    ):
        """Test status endpoint when dataset is not cached."""
        response = client.get("/api/embeddings/status/uncached_dataset")
        assert response.status_code == 200

        data = response.json()
        assert data["dataset_name"] == "uncached_dataset"
        assert data["cached"] is False
        assert data["num_embeddings"] == 0

    def test_embedding_status_cached(self, client, patched_cache):
        """Test status endpoint when dataset is cached."""
        response = client.get("/api/embeddings/status/test_dataset")
        assert response.status_code == 200

        data = response.json()
        assert data["dataset_name"] == "test_dataset"
        assert data["cached"] is True
        assert data["num_embeddings"] == 50
        assert data["embedding_dim"] == 768
        assert data["cached_at"] is not None
        assert data["expires_in_sec"] is not None

    def test_embedding_status_signrep_availability(self, client, force_signrep_unavailable):
        """Test that status endpoint reports SignRep availability."""
        response = client.get("/api/embeddings/status/any_dataset")
        assert response.status_code == 200

        data = response.json()
        # SignRep should not be available when forced unavailable
        assert "signrep_available" in data
        assert "signrep_loaded" in data
        # With force_signrep_unavailable, signrep should not be available
        assert data["signrep_available"] is False


class TestComputeEmbeddingsEndpoint:
    """Tests for POST /api/embeddings/compute."""

    def test_compute_embeddings_returns_cached(self, client, patched_cache):
        """Test that compute endpoint returns cached results if available."""
        response = client.post(
            "/api/embeddings/compute",
            json={"dataset_name": "test_dataset", "max_samples": 50},
        )
        assert response.status_code == 200

        data = response.json()
        assert data["dataset_name"] == "test_dataset"
        assert data["cached"] is True
        assert data["num_embeddings"] == 50

    def test_compute_embeddings_force_recompute_without_signrep(
        self, client, patched_cache, force_signrep_unavailable, clear_embedding_cache
    ):
        """Test force recompute fails gracefully without SignRep."""
        response = client.post(
            "/api/embeddings/compute",
            json={
                "dataset_name": "test_dataset",
                "max_samples": 10,
                "force_recompute": True,
            },
        )
        # Should fail since SignRep is not available
        assert response.status_code == 503

    def test_compute_embeddings_invalid_max_samples(self, client):
        """Test that invalid max_samples is rejected."""
        # Too high
        response = client.post(
            "/api/embeddings/compute",
            json={"dataset_name": "test", "max_samples": 10000},
        )
        assert response.status_code == 422  # Validation error

        # Too low
        response = client.post(
            "/api/embeddings/compute",
            json={"dataset_name": "test", "max_samples": 0},
        )
        assert response.status_code == 422


class TestVisualizeEmbeddingsEndpoint:
    """Tests for POST /api/embeddings/visualize."""

    def test_visualize_requires_cached_embeddings(self, client, clear_embedding_cache):
        """Test that visualize fails without cached embeddings."""
        response = client.post(
            "/api/embeddings/visualize",
            json={"dataset_name": "uncached_dataset"},
        )
        assert response.status_code == 400
        assert "No embeddings cached" in response.json()["detail"]

    def test_visualize_pca_returns_explained_variance(self, client, patched_cache):
        """Test that PCA projection includes explained variance."""
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "projection_method": "pca",
                "n_components": 2,
            },
        )
        assert response.status_code == 200

        data = response.json()
        assert data["projection_method"] == "pca"
        assert data["explained_variance"] is not None
        assert len(data["explained_variance"]) == 2
        # Explained variance should be between 0 and 1
        for ev in data["explained_variance"]:
            assert 0 <= ev <= 1

    def test_visualize_pca_3d(self, client, patched_cache):
        """Test 3D PCA projection."""
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "projection_method": "pca",
                "n_components": 3,
            },
        )
        assert response.status_code == 200

        data = response.json()
        assert len(data["explained_variance"]) == 3
        # All points should have z coordinate
        for point in data["points"]:
            assert point["z"] is not None

    def test_visualize_tsne_perplexity_auto_adjusted(self, client):
        """Test that t-SNE perplexity is auto-adjusted when > n_samples."""
        from sltk.api.routers import embeddings as emb_router

        # Create small dataset (less than default perplexity of 30)
        np.random.seed(42)
        small_embeddings = np.random.randn(15, 768).astype(np.float32)
        small_metadata = {
            "sample_ids": [f"s_{i}" for i in range(15)],
            "glosses": [f"G_{i}" for i in range(15)],
            "signer_ids": ["signer_0"] * 15,
            "video_paths": [f"/path/{i}.mp4" for i in range(15)],
        }
        emb_router._cache_embeddings("small_dataset", small_embeddings, small_metadata)

        try:
            response = client.post(
                "/api/embeddings/visualize",
                json={
                    "dataset_name": "small_dataset",
                    "projection_method": "tsne",
                    "perplexity": 50,  # Larger than n_samples
                },
            )
            assert response.status_code == 200
            # Should succeed with auto-adjusted perplexity
            data = response.json()
            assert data["num_points"] == 15
        finally:
            # Cleanup
            with emb_router._embedding_cache_lock:
                if "small_dataset" in emb_router._embedding_cache:
                    del emb_router._embedding_cache["small_dataset"]

    def test_visualize_max_samples_limit(self, client):
        """Test that max_samples limit is enforced."""
        from sltk.api.routers import embeddings as emb_router

        # Create larger dataset
        np.random.seed(42)
        large_embeddings = np.random.randn(100, 768).astype(np.float32)
        large_metadata = {
            "sample_ids": [f"s_{i}" for i in range(100)],
            "glosses": [f"G_{i % 10}" for i in range(100)],
            "signer_ids": [f"signer_{i % 5}" for i in range(100)],
            "video_paths": [f"/path/{i}.mp4" for i in range(100)],
        }
        emb_router._cache_embeddings("large_dataset", large_embeddings, large_metadata)

        try:
            response = client.post(
                "/api/embeddings/visualize",
                json={
                    "dataset_name": "large_dataset",
                    "max_samples": 30,
                },
            )
            assert response.status_code == 200
            data = response.json()
            assert data["num_points"] == 30  # Should be limited
        finally:
            # Cleanup
            with emb_router._embedding_cache_lock:
                if "large_dataset" in emb_router._embedding_cache:
                    del emb_router._embedding_cache["large_dataset"]

    def test_visualize_point_structure(self, client, patched_cache):
        """Test that visualization points have correct structure."""
        response = client.post(
            "/api/embeddings/visualize",
            json={"dataset_name": "test_dataset"},
        )
        assert response.status_code == 200

        data = response.json()
        assert len(data["points"]) > 0

        point = data["points"][0]
        assert "sample_id" in point
        assert "x" in point
        assert "y" in point
        assert "gloss" in point
        assert "signer_id" in point
        assert "cluster_id" in point
        assert "label" in point

    def test_visualize_returns_unique_labels(self, client, patched_cache):
        """Test that unique labels are returned."""
        response = client.post(
            "/api/embeddings/visualize",
            json={"dataset_name": "test_dataset"},
        )
        assert response.status_code == 200

        data = response.json()
        assert "unique_labels" in data
        assert isinstance(data["unique_labels"], list)

    def test_visualize_returns_statistics(self, client, patched_cache):
        """Test that statistics are returned."""
        response = client.post(
            "/api/embeddings/visualize",
            json={"dataset_name": "test_dataset"},
        )
        assert response.status_code == 200

        data = response.json()
        assert "statistics" in data
        assert "total_samples" in data["statistics"]


class TestClusteringEndpoints:
    """Tests for clustering in visualize endpoint."""

    def test_kmeans_clustering(self, client, patched_cache):
        """Test k-means clustering."""
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "cluster_method": "kmeans",
                "n_clusters": 5,
            },
        )
        assert response.status_code == 200

        data = response.json()
        assert "num_clusters" in data["statistics"]
        assert data["statistics"]["num_clusters"] == 5

        # All points should have cluster_id
        cluster_ids = {p["cluster_id"] for p in data["points"]}
        assert len(cluster_ids) == 5

    def test_kmeans_requires_n_clusters(self, client, patched_cache):
        """Test that k-means requires n_clusters parameter."""
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "cluster_method": "kmeans",
                # No n_clusters
            },
        )
        assert response.status_code == 400
        assert "n_clusters required" in response.json()["detail"]

    def test_hdbscan_clustering(self, client, patched_cache):
        """Test HDBSCAN clustering (if available)."""
        try:
            import hdbscan  # noqa: F401
        except ImportError:
            pytest.skip("hdbscan not installed")

        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "cluster_method": "hdbscan",
            },
        )
        assert response.status_code == 200

        data = response.json()
        assert "num_clusters" in data["statistics"]
        assert "num_noise_points" in data["statistics"]

    def test_no_clustering(self, client, patched_cache):
        """Test visualization without clustering."""
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "cluster_method": "none",
            },
        )
        assert response.status_code == 200

        data = response.json()
        # All points should have cluster_id = -1
        for point in data["points"]:
            assert point["cluster_id"] == -1


class TestClearCacheEndpoint:
    """Tests for DELETE /api/embeddings/cache/{dataset}."""

    def test_clear_cache_existing(self, client, patched_cache):
        """Test clearing existing cache entry."""
        response = client.delete("/api/embeddings/cache/test_dataset")
        assert response.status_code == 200

        data = response.json()
        assert data["status"] == "cleared"
        assert data["dataset_name"] == "test_dataset"

    def test_clear_cache_nonexistent(self, client, clear_embedding_cache):
        """Test clearing non-existent cache entry."""
        response = client.delete("/api/embeddings/cache/nonexistent")
        assert response.status_code == 200

        data = response.json()
        assert data["status"] == "not_cached"

    def test_clear_all_cache(self, client, patched_cache):
        """Test clearing all cache entries."""
        # First verify cache has entries
        from sltk.api.routers import embeddings as emb_router

        assert len(emb_router._embedding_cache) > 0

        response = client.delete("/api/embeddings/cache")
        assert response.status_code == 200

        data = response.json()
        assert data["status"] == "cleared"
        assert data["datasets_cleared"] >= 1


class TestSignRepStatusEndpoint:
    """Tests for GET /api/embeddings/signrep/status."""

    def test_signrep_status_unavailable(self, client, force_signrep_unavailable):
        """Test SignRep status when not available."""
        response = client.get("/api/embeddings/signrep/status")
        assert response.status_code == 200

        data = response.json()
        assert data["available"] is False
        assert data["loaded"] is False
        assert data["error"] is not None

    def test_signrep_status_available(self, client, clear_signrep_state, mock_signrep_embedder):
        """Test SignRep status when available."""
        from sltk.api.routers import embeddings as emb_router

        # Set up mocked embedder
        emb_router._signrep_embedder = mock_signrep_embedder
        emb_router._signrep_init_error = None

        try:
            response = client.get("/api/embeddings/signrep/status")
            assert response.status_code == 200

            data = response.json()
            assert data["available"] is True
            assert data["loaded"] is True
            assert data["embedding_dim"] == 768
            assert data["error"] is None
        finally:
            # Cleanup
            emb_router._signrep_embedder = None


# ============================================================================
# Error handling tests
# ============================================================================


class TestErrorHandling:
    """Tests for error handling in embedding endpoints."""

    def test_signrep_unavailable_returns_503(
        self, client, force_signrep_unavailable, clear_embedding_cache
    ):
        """Test that 503 is returned when SignRep is unavailable."""
        response = client.post(
            "/api/embeddings/compute",
            json={
                "dataset_name": "test",
                "max_samples": 10,
                "force_recompute": True,
            },
        )
        assert response.status_code == 503
        assert "SignRep" in response.json()["detail"]

    def test_invalid_projection_method(self, client, patched_cache):
        """Test that invalid projection method is rejected."""
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "projection_method": "invalid_method",
            },
        )
        assert response.status_code == 422  # Validation error

    def test_invalid_cluster_method(self, client, patched_cache):
        """Test that invalid cluster method is rejected."""
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "cluster_method": "invalid_method",
            },
        )
        assert response.status_code == 422  # Validation error

    def test_invalid_n_components(self, client, patched_cache):
        """Test that invalid n_components is rejected."""
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "n_components": 4,  # Only 2 or 3 allowed
            },
        )
        assert response.status_code == 422  # Validation error

    def test_perplexity_out_of_range(self, client, patched_cache):
        """Test that perplexity out of range is rejected."""
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "test_dataset",
                "projection_method": "tsne",
                "perplexity": 100,  # Max is 50
            },
        )
        assert response.status_code == 422  # Validation error


# ============================================================================
# Dimensionality reduction helper tests
# ============================================================================


class TestDimensionalityReduction:
    """Tests for dimensionality reduction helper functions."""

    def test_pca_reduction(self, mock_embeddings):
        """Test PCA dimensionality reduction."""
        from sltk.api.routers.embeddings import _reduce_dimensions_sync

        projections, explained_variance = _reduce_dimensions_sync(mock_embeddings, "pca", 2)

        assert projections.shape == (50, 2)
        assert explained_variance is not None
        assert len(explained_variance) == 2
        assert sum(explained_variance) <= 1.0

    def test_tsne_reduction(self, mock_embeddings):
        """Test t-SNE dimensionality reduction."""
        from sltk.api.routers.embeddings import _reduce_dimensions_sync

        projections, explained_variance = _reduce_dimensions_sync(
            mock_embeddings[:30], "tsne", 2, perplexity=10
        )

        assert projections.shape == (30, 2)
        assert explained_variance is None  # t-SNE doesn't provide this

    def test_umap_reduction(self, mock_embeddings):
        """Test UMAP dimensionality reduction."""
        try:
            import umap  # noqa: F401
        except ImportError:
            pytest.skip("umap-learn not installed")

        from sltk.api.routers.embeddings import _reduce_dimensions_sync

        projections, explained_variance = _reduce_dimensions_sync(
            mock_embeddings, "umap", 2, n_neighbors=10
        )

        assert projections.shape == (50, 2)
        assert explained_variance is None  # UMAP doesn't provide this


class TestClustering:
    """Tests for clustering helper functions."""

    def test_kmeans_clustering_sync(self, mock_embeddings):
        """Test k-means clustering."""
        from sltk.api.routers.embeddings import _cluster_sync

        cluster_ids = _cluster_sync(mock_embeddings, "kmeans", n_clusters=5)

        assert len(cluster_ids) == 50
        assert set(cluster_ids) == {0, 1, 2, 3, 4}

    def test_hdbscan_clustering_sync(self, mock_embeddings):
        """Test HDBSCAN clustering."""
        try:
            import hdbscan  # noqa: F401
        except ImportError:
            pytest.skip("hdbscan not installed")

        from sltk.api.routers.embeddings import _cluster_sync

        cluster_ids = _cluster_sync(mock_embeddings, "hdbscan")

        assert len(cluster_ids) == 50
        # HDBSCAN may have -1 for noise points

    def test_no_clustering_sync(self, mock_embeddings):
        """Test no clustering."""
        from sltk.api.routers.embeddings import _cluster_sync

        cluster_ids = _cluster_sync(mock_embeddings, "none")

        assert len(cluster_ids) == 50
        assert all(c == -1 for c in cluster_ids)


# ============================================================================
# Pydantic model validation tests
# ============================================================================


class TestPydanticModels:
    """Tests for Pydantic model validation."""

    def test_compute_embeddings_request_defaults(self):
        """Test ComputeEmbeddingsRequest defaults."""
        from sltk.api.routers.embeddings import ComputeEmbeddingsRequest

        req = ComputeEmbeddingsRequest(dataset_name="test")
        assert req.max_samples == 500
        assert req.aggregation == "mean"
        assert req.force_recompute is False

    def test_visualize_embeddings_request_defaults(self):
        """Test VisualizeEmbeddingsRequest defaults."""
        from sltk.api.routers.embeddings import VisualizeEmbeddingsRequest

        req = VisualizeEmbeddingsRequest(dataset_name="test")
        assert req.projection_method == "pca"
        assert req.n_components == 2
        assert req.cluster_method == "none"
        assert req.max_samples == 500
        assert req.n_neighbors == 15
        assert req.min_dist == 0.1

    def test_embedding_point_model(self):
        """Test EmbeddingPoint model."""
        from sltk.api.routers.embeddings import EmbeddingPoint

        point = EmbeddingPoint(
            sample_id="test_001",
            x=1.0,
            y=2.0,
            z=3.0,
            gloss="TEST",
            signer_id="signer_01",
            cluster_id=0,
            label="TEST",
        )
        assert point.sample_id == "test_001"
        assert point.z == 3.0

    def test_embedding_status_response_model(self):
        """Test EmbeddingStatusResponse model."""
        from sltk.api.routers.embeddings import EmbeddingStatusResponse

        resp = EmbeddingStatusResponse(
            dataset_name="test",
            cached=True,
            num_embeddings=100,
            embedding_dim=768,
        )
        assert resp.cached is True
        assert resp.expires_in_sec is None  # Optional


# ============================================================================
# Integration tests
# ============================================================================


class TestIntegration:
    """Integration tests for the embedding workflow."""

    def test_full_workflow_with_mock(self, client, clear_embedding_cache):
        """Test the full embedding workflow with mocked components."""
        from sltk.api.routers import embeddings as emb_router

        # Step 1: Check status (should be empty)
        response = client.get("/api/embeddings/status/workflow_test")
        assert response.status_code == 200
        assert response.json()["cached"] is False

        # Step 2: Manually cache embeddings (simulating compute)
        np.random.seed(42)
        embeddings = np.random.randn(20, 768).astype(np.float32)
        metadata = {
            "sample_ids": [f"s_{i}" for i in range(20)],
            "glosses": [f"G_{i % 5}" for i in range(20)],
            "signer_ids": [f"signer_{i % 3}" for i in range(20)],
            "video_paths": [f"/path/{i}.mp4" for i in range(20)],
        }
        emb_router._cache_embeddings("workflow_test", embeddings, metadata)

        # Step 3: Check status again (should be cached)
        response = client.get("/api/embeddings/status/workflow_test")
        assert response.status_code == 200
        assert response.json()["cached"] is True

        # Step 4: Visualize with PCA
        response = client.post(
            "/api/embeddings/visualize",
            json={
                "dataset_name": "workflow_test",
                "projection_method": "pca",
                "cluster_method": "kmeans",
                "n_clusters": 3,
            },
        )
        assert response.status_code == 200
        data = response.json()
        assert data["num_points"] == 20
        assert data["statistics"]["num_clusters"] == 3

        # Step 5: Clear cache
        response = client.delete("/api/embeddings/cache/workflow_test")
        assert response.status_code == 200
        assert response.json()["status"] == "cleared"

        # Step 6: Verify cache is cleared
        response = client.get("/api/embeddings/status/workflow_test")
        assert response.status_code == 200
        assert response.json()["cached"] is False

    def test_compute_returns_cached_without_recompute(self, client, patched_cache):
        """Test that compute endpoint returns cached without recomputing."""
        response = client.post(
            "/api/embeddings/compute",
            json={"dataset_name": "test_dataset"},
        )
        assert response.status_code == 200

        data = response.json()
        assert data["cached"] is True
        assert data["compute_time_sec"] is None  # No computation



"""
Search endpoints for SLTK API.

Provides efficient cross-dataset gloss and transcription search.
"""

from __future__ import annotations

import asyncio
import logging
import re
import time
from typing import Any

from fastapi import APIRouter, Query, Request

from sltk.api.dependencies import get_cached_dataset, rate_limit

router = APIRouter(prefix="/api/search", tags=["search"])
logger = logging.getLogger(__name__)

# Pre-compiled patterns for text normalization
_WORD_SPLIT_PATTERN = re.compile(r"[\s\-_/]+")
_NON_ALPHA_PATTERN = re.compile(r"[^a-z0-9]")


def _normalize_text(text: str) -> str:
    """Normalize text for comparison."""
    return _NON_ALPHA_PATTERN.sub("", text.lower())


def _levenshtein_distance(s1: str, s2: str) -> int:
    """
    Compute Levenshtein distance between two strings.

    Simple implementation optimized for short strings (glosses).
    """
    if len(s1) < len(s2):
        s1, s2 = s2, s1

    if len(s2) == 0:
        return len(s1)

    previous_row = list(range(len(s2) + 1))

    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def _similarity_score(s1: str, s2: str) -> float:
    """
    Compute similarity score between two strings (0-1).

    Uses normalized Levenshtein distance.
    """
    s1_norm = _normalize_text(s1)
    s2_norm = _normalize_text(s2)

    if not s1_norm or not s2_norm:
        return 0.0

    if s1_norm == s2_norm:
        return 1.0

    # Check for prefix/suffix match (common in glosses)
    if s1_norm.startswith(s2_norm) or s2_norm.startswith(s1_norm):
        return 0.9

    max_len = max(len(s1_norm), len(s2_norm))
    distance = _levenshtein_distance(s1_norm, s2_norm)
    return 1.0 - (distance / max_len)


def _find_matches_in_text(
    query: str,
    text: str,
    min_similarity: float = 0.7,
) -> list[tuple[str, float]]:
    """
    Find fuzzy matches for query in text.

    Returns list of (matched_word, similarity_score) tuples.
    """
    if not text:
        return []

    matches = []
    query_norm = _normalize_text(query)

    if not query_norm:
        return []

    # Split text into words
    words = _WORD_SPLIT_PATTERN.split(text)

    for word in words:
        if not word:
            continue

        score = _similarity_score(query, word)
        if score >= min_similarity:
            matches.append((word, score))

    return matches


# ============================================================================
# Search Result Models
# ============================================================================

from pydantic import BaseModel, Field


class SearchResult(BaseModel):
    """A single search result."""

    dataset: str
    sample_id: str
    match_type: str  # 'exact_gloss', 'fuzzy_gloss', 'transcription'
    matched_text: str
    similarity: float
    start: float
    end: float
    context: list[str] = Field(default_factory=list)
    video_path: str | None = None
    signer_id: str | None = None
    tier: str | None = None


class SearchResponse(BaseModel):
    """Response for gloss search."""

    query: str
    total_results: int
    datasets_searched: int
    results: list[SearchResult]
    search_time_ms: float


# ============================================================================
# Search Endpoints
# ============================================================================


@router.get("/glosses", response_model=SearchResponse)
@rate_limit("30/minute")
async def search_glosses(
    request: Request,
    q: str = Query(..., min_length=1, max_length=100, description="Search query"),
    datasets: str | None = Query(None, description="Comma-separated dataset names (None = all)"),
    include_transcriptions: bool = Query(
        True, description="Search transcriptions for similar words"
    ),
    min_similarity: float = Query(
        0.7, ge=0.5, le=1.0, description="Minimum similarity for fuzzy matching"
    ),
    limit: int = Query(50, ge=1, le=200, description="Maximum results to return"),
    max_samples_per_dataset: int = Query(
        200, ge=1, le=1000, description="Max samples to search per dataset"
    ),
):
    """
    Search for glosses across all connected datasets.

    Features:
    - Exact gloss matching (highest priority)
    - Fuzzy gloss matching (for typos/variants)
    - Transcription search (finds similar English words)
    - Returns results sorted by relevance

    The search is optimized for quick response times by:
    - Limiting samples searched per dataset
    - Using efficient string matching algorithms
    - Running dataset searches in parallel
    """
    start_time = time.time()
    query = q.strip()
    query_lower = query.lower()

    # Determine which datasets to search
    from pathlib import Path

    from sltk.config import DATASET_PATHS
    from sltk.data.datasets import DATASETS

    if datasets:
        dataset_names = [d.strip() for d in datasets.split(",") if d.strip()]
    else:
        # Get all accessible datasets (quick check - only verify root exists)
        dataset_names = []
        for name in DATASETS.keys():
            dataset_info = DATASET_PATHS.get(name, {})
            root_path = dataset_info.get("root")
            if root_path:
                try:
                    path = Path(root_path)
                    # Quick existence check without listing contents
                    if path.exists() and path.is_dir():
                        dataset_names.append(name)
                except (OSError, PermissionError):
                    pass

    if not dataset_names:
        return SearchResponse(
            query=query,
            total_results=0,
            datasets_searched=0,
            results=[],
            search_time_ms=0,
        )

    # Search function for a single dataset
    def _search_dataset(name: str) -> list[SearchResult]:
        """Search a single dataset for matches."""
        results: list[SearchResult] = []
        max_results_per_dataset = limit  # Stop searching once we have enough

        try:
            dataset = get_cached_dataset(name, split="all")
        except Exception as e:
            logger.debug(f"Could not load dataset {name}: {e}")
            return []

        samples_searched = 0

        for sample_id in dataset._index[:max_samples_per_dataset]:
            # Early termination if we have enough results
            if len(results) >= max_results_per_dataset:
                break

            samples_searched += 1

            try:
                sample = dataset._load_sample(sample_id)
            except Exception:
                continue

            video_path = str(sample.video_path) if sample.video_path else None
            signer_id = sample.signer_id

            # Search in glosses (exact and fuzzy)
            if sample.glosses:
                for i, gloss in enumerate(sample.glosses):
                    if not gloss:
                        continue

                    gloss_lower = gloss.lower()

                    # Exact match
                    if gloss_lower == query_lower:
                        context_before = sample.glosses[max(0, i - 2) : i]
                        context_after = sample.glosses[i + 1 : i + 3]

                        results.append(
                            SearchResult(
                                dataset=name,
                                sample_id=sample.id,
                                match_type="exact_gloss",
                                matched_text=gloss,
                                similarity=1.0,
                                start=0.0,  # Gloss-level doesn't have timing
                                end=0.0,
                                context=context_before + context_after,
                                video_path=video_path,
                                signer_id=signer_id,
                            )
                        )
                    # Fuzzy match
                    elif min_similarity < 1.0:
                        score = _similarity_score(query, gloss)
                        if score >= min_similarity and score < 1.0:
                            context_before = sample.glosses[max(0, i - 2) : i]
                            context_after = sample.glosses[i + 1 : i + 3]

                            results.append(
                                SearchResult(
                                    dataset=name,
                                    sample_id=sample.id,
                                    match_type="fuzzy_gloss",
                                    matched_text=gloss,
                                    similarity=score,
                                    start=0.0,
                                    end=0.0,
                                    context=context_before + context_after,
                                    video_path=video_path,
                                    signer_id=signer_id,
                                )
                            )

            # Search in segments (exact and fuzzy)
            if sample.segments:
                for i, seg in enumerate(sample.segments):
                    label = seg.label
                    if not label:
                        continue

                    label_lower = label.lower()

                    # Exact match
                    if label_lower == query_lower:
                        context = []
                        for j in range(max(0, i - 2), min(len(sample.segments), i + 3)):
                            if j != i and sample.segments[j].label:
                                context.append(sample.segments[j].label)

                        results.append(
                            SearchResult(
                                dataset=name,
                                sample_id=sample.id,
                                match_type="exact_gloss",
                                matched_text=label,
                                similarity=1.0,
                                start=seg.start,
                                end=seg.end,
                                context=context[:4],
                                video_path=video_path,
                                signer_id=signer_id,
                                tier=seg.tier,
                            )
                        )
                    # Fuzzy match
                    elif min_similarity < 1.0:
                        score = _similarity_score(query, label)
                        if score >= min_similarity and score < 1.0:
                            context = []
                            for j in range(max(0, i - 2), min(len(sample.segments), i + 3)):
                                if j != i and sample.segments[j].label:
                                    context.append(sample.segments[j].label)

                            results.append(
                                SearchResult(
                                    dataset=name,
                                    sample_id=sample.id,
                                    match_type="fuzzy_gloss",
                                    matched_text=label,
                                    similarity=score,
                                    start=seg.start,
                                    end=seg.end,
                                    context=context[:4],
                                    video_path=video_path,
                                    signer_id=signer_id,
                                    tier=seg.tier,
                                )
                            )

            # Search in transcriptions (text field)
            if include_transcriptions and hasattr(sample, "text") and sample.text:
                matches = _find_matches_in_text(query, sample.text, min_similarity)
                for matched_word, score in matches:
                    results.append(
                        SearchResult(
                            dataset=name,
                            sample_id=sample.id,
                            match_type="transcription",
                            matched_text=matched_word,
                            similarity=score,
                            start=0.0,
                            end=0.0,
                            context=[
                                sample.text[:100] + "..." if len(sample.text) > 100 else sample.text
                            ],
                            video_path=video_path,
                            signer_id=signer_id,
                        )
                    )

            # Also search in metadata if it contains transcription-like fields
            if include_transcriptions and sample.metadata:
                for key in ["translation", "transcription", "subtitle", "caption", "english"]:
                    if key in sample.metadata and isinstance(sample.metadata[key], str):
                        text = sample.metadata[key]
                        matches = _find_matches_in_text(query, text, min_similarity)
                        for matched_word, score in matches:
                            results.append(
                                SearchResult(
                                    dataset=name,
                                    sample_id=sample.id,
                                    match_type="transcription",
                                    matched_text=matched_word,
                                    similarity=score,
                                    start=0.0,
                                    end=0.0,
                                    context=[text[:100] + "..." if len(text) > 100 else text],
                                    video_path=video_path,
                                    signer_id=signer_id,
                                )
                            )

        return results

    # Run searches in parallel
    tasks = [asyncio.to_thread(_search_dataset, name) for name in dataset_names]

    all_results_nested = await asyncio.gather(*tasks, return_exceptions=True)

    # Flatten and filter results
    all_results: list[SearchResult] = []
    datasets_searched = 0

    for result in all_results_nested:
        if isinstance(result, BaseException):
            logger.debug(f"Search error: {result}")
            continue
        datasets_searched += 1
        all_results.extend(result)

    # Sort by similarity (descending) and match type priority
    type_priority = {"exact_gloss": 0, "fuzzy_gloss": 1, "transcription": 2}
    all_results.sort(key=lambda r: (type_priority.get(r.match_type, 3), -r.similarity))

    # Deduplicate (same dataset + sample + matched_text)
    seen = set()
    unique_results = []
    for r in all_results:
        key = (r.dataset, r.sample_id, r.matched_text.lower())
        if key not in seen:
            seen.add(key)
            unique_results.append(r)

    # Limit results
    final_results = unique_results[:limit]

    search_time_ms = (time.time() - start_time) * 1000

    return SearchResponse(
        query=query,
        total_results=len(unique_results),
        datasets_searched=datasets_searched,
        results=final_results,
        search_time_ms=round(search_time_ms, 2),
    )


@router.get("/suggestions")
async def get_search_suggestions(
    q: str = Query(..., min_length=1, max_length=50, description="Partial query"),
    limit: int = Query(10, ge=1, le=20, description="Max suggestions"),
):
    """
    Get autocomplete suggestions based on vocabulary from cached datasets.

    Uses vocabulary cache if available for fast response.
    """
    from sltk.api.dependencies import get_vocabulary_cache

    query_lower = q.lower().strip()
    suggestions: set[tuple[str, Any]] = set()

    _vocabulary_cache, _vocabulary_cache_lock, _ = get_vocabulary_cache()

    with _vocabulary_cache_lock:
        for cache_key, (cached_data, _) in _vocabulary_cache.items():
            if len(suggestions) >= limit * 2:  # Get extras for filtering
                break

            frequencies = cached_data.get("frequencies", {})
            for gloss, freq in frequencies.items():
                if gloss.lower().startswith(query_lower):
                    suggestions.add((gloss, freq))
                elif query_lower in gloss.lower():
                    suggestions.add((gloss, freq))

    # Sort by frequency and take top results
    sorted_suggestions = sorted(suggestions, key=lambda x: -x[1])[:limit]

    return {
        "query": q,
        "suggestions": [s[0] for s in sorted_suggestions],
    }



"""Tests for rotation conversion utilities."""

import numpy as np

from sltk.data.rotations import (
    RotationConverter,
    axis_angle_to_matrix,
    matrix_to_6d,
    matrix_to_axis_angle,
    rotation_6d_to_matrix,
)


class TestAxisAngleToMatrix:
    """Tests for axis-angle to rotation matrix conversion."""

    def test_identity(self):
        """Zero axis-angle should give identity matrix."""
        aa = np.array([0.0, 0.0, 0.0])
        mat = axis_angle_to_matrix(aa)

        expected = np.eye(3)
        np.testing.assert_array_almost_equal(mat, expected, decimal=5)

    def test_90_degree_rotation(self):
        """Test 90 degree rotation around z-axis."""
        # 90 degrees = pi/2 radians around z-axis
        aa = np.array([0.0, 0.0, np.pi / 2])
        mat = axis_angle_to_matrix(aa)

        # Expected: rotate x to y, y to -x
        expected = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]], dtype=np.float64)
        np.testing.assert_array_almost_equal(mat, expected, decimal=5)

    def test_batch(self):
        """Test batch conversion."""
        aa = np.random.randn(10, 3)
        mat = axis_angle_to_matrix(aa)

        assert mat.shape == (10, 3, 3)

    def test_roundtrip(self):
        """Test axis-angle -> matrix -> axis-angle roundtrip."""
        aa_original = np.array([0.1, 0.2, 0.3])
        mat = axis_angle_to_matrix(aa_original)
        aa_recovered = matrix_to_axis_angle(mat)

        # Should recover original (up to sign ambiguity for pi rotations)
        np.testing.assert_array_almost_equal(aa_original, aa_recovered, decimal=5)


class TestMatrixTo6D:
    """Tests for rotation matrix to 6D conversion."""

    def test_identity(self):
        """Identity matrix should give [1,0,0, 0,1,0]."""
        mat = np.eye(3)
        r6d = matrix_to_6d(mat)

        expected = np.array([1, 0, 0, 0, 1, 0])
        np.testing.assert_array_almost_equal(r6d, expected, decimal=5)

    def test_roundtrip(self):
        """Test matrix -> 6D -> matrix roundtrip."""
        # Create a valid rotation matrix
        aa = np.array([0.1, 0.2, 0.3])
        mat_original = axis_angle_to_matrix(aa)

        r6d = matrix_to_6d(mat_original)
        mat_recovered = rotation_6d_to_matrix(r6d)

        np.testing.assert_array_almost_equal(mat_original, mat_recovered, decimal=5)


class TestRotation6DToMatrix:
    """Tests for 6D to rotation matrix conversion."""

    def test_orthogonalization(self):
        """Test that output is a valid rotation matrix."""
        # Input slightly non-orthogonal 6D
        r6d = np.array([1.01, 0.01, 0.01, 0.01, 0.99, 0.01])
        mat = rotation_6d_to_matrix(r6d)

        # Check orthogonality: R^T @ R = I
        rtrt = mat.T @ mat
        np.testing.assert_array_almost_equal(rtrt, np.eye(3), decimal=5)

        # Check determinant = 1 (proper rotation, not reflection)
        det = np.linalg.det(mat)
        np.testing.assert_almost_equal(det, 1.0, decimal=5)


class TestRotationConverter:
    """Tests for RotationConverter class."""

    def test_to_6d(self):
        """Test conversion to 6D format."""
        converter = RotationConverter(target_format="rotation_6d")

        aa = np.array([0.1, 0.2, 0.3])
        r6d = converter.from_axis_angle(aa)

        assert r6d.shape == (6,)

    def test_to_matrix(self):
        """Test conversion to matrix format."""
        converter = RotationConverter(target_format="rotation_matrix")

        aa = np.array([0.1, 0.2, 0.3])
        mat = converter.from_axis_angle(aa)

        assert mat.shape == (3, 3)

    def test_convert_method(self):
        """Test the convert method."""
        converter = RotationConverter(target_format="rotation_6d")

        aa = np.array([0.1, 0.2, 0.3])
        r6d = converter.convert(aa, source_format="axis_angle")

        # Should be same as from_axis_angle
        expected = converter.from_axis_angle(aa)
        np.testing.assert_array_almost_equal(r6d, expected)



"""
API tests for the segmentation v2 router (/api/segmentation/*).

Tests:
- GET /api/segmentation/status -- checkpoint availability check
- POST /api/segmentation/segment -- single H5 file segmentation (mocked runner)
- POST /api/segmentation/segment/batch -- batch directory (mocked runner)
- Error cases: missing file, bad extension, model unavailable

Run with:
    pytest tests/test_segmentation_api.py -v
"""

from __future__ import annotations

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

torch = pytest.importorskip("torch")
from fastapi.testclient import TestClient


@pytest.fixture()
def client():
    """Create a test client for the SLTK API."""
    from sltk.api.main import app

    return TestClient(app)


# ===========================================================================
# GET /api/segmentation/status
# ===========================================================================


class TestSegmentationStatus:
    """Test the status endpoint."""

    def test_status_returns_200(self, client):
        resp = client.get("/api/segmentation/status")
        assert resp.status_code == 200
        data = resp.json()
        assert "checkpoint_exists" in data
        assert "gpu_available" in data

    def test_status_reports_checkpoint(self, client):
        resp = client.get("/api/segmentation/status")
        data = resp.json()
        # checkpoint_exists is bool based on whether file is on disk
        assert isinstance(data["checkpoint_exists"], bool)


# ===========================================================================
# POST /api/segmentation/segment
# ===========================================================================


class TestSegmentSingle:
    """Test single H5 file segmentation endpoint."""

    def test_segment_with_mocked_runner(self, client, tmp_path):
        """Mock the segmentation pipeline and verify endpoint plumbing."""
        # Create a fake H5 file
        h5_file = tmp_path / "test_video.h5"
        h5_file.touch()

        fake_features = torch.randn(100, 192)
        fake_labels = torch.tensor([0] * 10 + [2] + [1] * 20 + [0] * 10 + [2] + [1] * 15 + [0] * 43)

        with (
            patch(
                "sltk.api.routers.segmentation.validate_path",
                return_value=h5_file,
            ),
            patch(
                "sltk.segmentation.h5_loader.h5_to_features",
                return_value=fake_features,
            ),
            patch(
                "sltk.segmentation.runner.get_runner",
            ) as mock_get_runner,
        ):
            mock_runner = MagicMock()
            mock_runner.predict.return_value = fake_labels
            mock_get_runner.return_value = mock_runner

            resp = client.post(
                "/api/segmentation/segment",
                json={"h5_path": str(h5_file)},
            )

        assert resp.status_code == 200
        data = resp.json()
        assert data["video_name"] == "test_video"
        assert data["num_frames"] == 100
        assert len(data["segments"]) == 2
        assert len(data["labels"]) == 100

    def test_segment_rejects_non_h5(self, client, tmp_path):
        """Non-.h5 files should be rejected."""
        bad_file = tmp_path / "video.mp4"
        bad_file.touch()

        with patch(
            "sltk.api.routers.segmentation.validate_path",
            return_value=bad_file,
        ):
            resp = client.post(
                "/api/segmentation/segment",
                json={"h5_path": str(bad_file)},
            )

        assert resp.status_code == 400
        assert "H5" in resp.json()["detail"]

    def test_segment_model_unavailable(self, client, tmp_path):
        """When model checkpoint is missing, return 503."""
        h5_file = tmp_path / "test.h5"
        h5_file.touch()

        with (
            patch(
                "sltk.api.routers.segmentation.validate_path",
                return_value=h5_file,
            ),
            patch(
                "sltk.segmentation.h5_loader.h5_to_features",
                return_value=torch.randn(50, 192),
            ),
            patch(
                "sltk.segmentation.runner.get_runner",
                side_effect=FileNotFoundError("No checkpoint"),
            ),
        ):
            resp = client.post(
                "/api/segmentation/segment",
                json={"h5_path": str(h5_file)},
            )

        assert resp.status_code == 503
        assert "unavailable" in resp.json()["detail"].lower()


# ===========================================================================
# POST /api/segmentation/segment/batch
# ===========================================================================


class TestSegmentBatch:
    """Test batch directory segmentation endpoint."""

    def test_batch_with_mocked_runner(self, client, tmp_path):
        """Mock the runner and verify batch endpoint plumbing."""
        h5_dir = tmp_path / "h5_files"
        h5_dir.mkdir()
        (h5_dir / "a.h5").touch()
        (h5_dir / "b.h5").touch()

        predictions = {
            "a": torch.tensor([0, 2, 1, 1, 0]),
            "b": torch.tensor([2, 1, 0, 0, 0]),
        }

        with (
            patch(
                "sltk.api.routers.segmentation.validate_path",
                return_value=h5_dir,
            ),
            patch(
                "sltk.segmentation.runner.get_runner",
            ) as mock_get_runner,
        ):
            mock_runner = MagicMock()
            mock_runner.run_h5.return_value = predictions
            mock_get_runner.return_value = mock_runner

            resp = client.post(
                "/api/segmentation/segment/batch",
                json={"directory": str(h5_dir)},
            )

        assert resp.status_code == 200
        data = resp.json()
        assert data["total_videos"] == 2
        assert "a" in data["results"]
        assert "b" in data["results"]

    def test_batch_rejects_non_directory(self, client, tmp_path):
        """Non-directory paths should be rejected."""
        regular_file = tmp_path / "not_a_dir.txt"
        regular_file.touch()

        with patch(
            "sltk.api.routers.segmentation.validate_path",
            return_value=regular_file,
        ):
            resp = client.post(
                "/api/segmentation/segment/batch",
                json={"directory": str(regular_file)},
            )

        assert resp.status_code == 400
        assert "directory" in resp.json()["detail"].lower()

    def test_batch_rejects_invalid_format(self, client, tmp_path):
        """Invalid output_format should be rejected."""
        h5_dir = tmp_path / "h5s"
        h5_dir.mkdir()

        with patch(
            "sltk.api.routers.segmentation.validate_path",
            return_value=h5_dir,
        ):
            resp = client.post(
                "/api/segmentation/segment/batch",
                json={"directory": str(h5_dir), "output_format": "csv"},
            )

        assert resp.status_code == 400
        assert "output_format" in resp.json()["detail"].lower()



import pickle

import smplx
import torch
from smplx.lbs import vertices2joints
from smplx.utils import MANOOutput, to_tensor
from smplx.vertex_ids import vertex_ids


# SECURITY: Restricted unpickler to prevent arbitrary code execution
# Only allows numpy arrays and basic types needed for joint regressor data
class _RestrictedUnpickler(pickle.Unpickler):
    """Unpickler that only allows safe types for MANO joint regressor."""

    ALLOWED_MODULES = {
        "numpy": {"ndarray", "dtype", "float64", "float32", "int64", "int32"},
        "numpy.core.multiarray": {"_reconstruct", "scalar"},
        "builtins": {"dict", "list", "tuple", "str", "bytes", "int", "float", "bool"},
        "_codecs": {"encode"},
    }

    def find_class(self, module, name):
        # Allow numpy reconstruction
        if module.startswith("numpy"):
            import numpy
            if hasattr(numpy, name):
                return getattr(numpy, name)
            parts = module.split(".")
            obj = numpy
            for part in parts[1:]:
                if hasattr(obj, part):
                    obj = getattr(obj, part)
            if hasattr(obj, name):
                return getattr(obj, name)

        if module in self.ALLOWED_MODULES:
            allowed = self.ALLOWED_MODULES[module]
            if name in allowed:
                return super().find_class(module, name)

        raise ValueError(
            f"Restricted unpickling: '{module}.{name}' is not allowed. "
            "Only numpy arrays and basic Python types are permitted."
        )


def _safe_load_pickle(filepath, encoding="latin1"):
    """Safely load a pickle file with restricted unpickling."""
    with open(filepath, "rb") as f:
        return _RestrictedUnpickler(f, encoding=encoding).load()


class MANO(smplx.MANOLayer):
    def __init__(self, *args, joint_regressor_extra: str | None = None, **kwargs):
        """
        Extension of the official MANO implementation to support more joints.
        Args:
            Same as MANOLayer.
            joint_regressor_extra (str): Path to extra joint regressor.
        """
        super().__init__(*args, **kwargs)
        mano_to_openpose = [
            0,
            13,
            14,
            15,
            16,
            1,
            2,
            3,
            17,
            4,
            5,
            6,
            18,
            10,
            11,
            12,
            19,
            7,
            8,
            9,
            20,
        ]

        # 2, 3, 5, 4, 1
        if joint_regressor_extra is not None:
            # SECURITY: Use restricted unpickler to prevent arbitrary code execution
            self.register_buffer(
                "joint_regressor_extra",
                torch.tensor(
                    _safe_load_pickle(joint_regressor_extra, encoding="latin1"),
                    dtype=torch.float32,
                ),
            )
        self.register_buffer(
            "extra_joints_idxs", to_tensor(list(vertex_ids["mano"].values()), dtype=torch.long)
        )
        self.register_buffer("joint_map", torch.tensor(mano_to_openpose, dtype=torch.long))

    def forward(self, *args, **kwargs) -> MANOOutput:
        """
        Run forward pass. Same as MANO and also append an extra set of joints if joint_regressor_extra is specified.
        """
        mano_output = super().forward(*args, **kwargs)
        extra_joints = torch.index_select(mano_output.vertices, 1, self.extra_joints_idxs)
        joints = torch.cat([mano_output.joints, extra_joints], dim=1)
        joints = joints[:, self.joint_map, :]
        if hasattr(self, "joint_regressor_extra"):
            extra_joints = vertices2joints(self.joint_regressor_extra, mano_output.vertices)
            joints = torch.cat([joints, extra_joints], dim=1)
        mano_output.joints = joints
        return mano_output
